/****************************************************************************
 * This file is part of the project AqFinance.
 * AqFinance (c) by 2007 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif



#include "aqdb_object_p.h"
#include <aqdatabase/aqdb_value.h>

#include <gwenhywfar/buffer.h>
#include <gwenhywfar/debug.h>
#include <gwenhywfar/gwentime.h>
#include <gwenhywfar/text.h>



GWEN_LIST_FUNCTIONS(AQDB_OBJECT, AQDB_Object)
GWEN_LIST2_FUNCTIONS(AQDB_OBJECT, AQDB_Object)






AQDB_OBJECT_FIELD *AQDB_ObjectField_new(const uint8_t *p, uint32_t len) {
  AQDB_OBJECT_FIELD *field;

  field=(AQDB_OBJECT_FIELD*) GWEN_Memory_malloc(sizeof(AQDB_OBJECT_FIELD)+len);
  assert(field);
  memset(field, 0, sizeof(AQDB_OBJECT_FIELD)+len);

  field->refCount=1;
  field->fieldLen=len;
  field->type=AQDB_DataType_None;

  if (len && p)
    memmove(field->dataBegin, p, len);

  return field;
}



void AQDB_ObjectField_free(AQDB_OBJECT_FIELD *field) {
  if (field) {
    assert(field->refCount);
    if (field->refCount==1) {
      AQDB_ObjectField_ClearTypedPointer(field);
      field->refCount=0;
      GWEN_Memory_dealloc((void*) field);
    }
    else {
      field->refCount--;
    }
  }
}



void AQDB_ObjectField_ClearTypedPointer(AQDB_OBJECT_FIELD *field) {
  assert(field);
  assert(field->refCount);
  if (field->typedPointer!=NULL) {
    switch(field->type) {
    case AQDB_DataType_Date:
      GWEN_Date_free((GWEN_DATE*)(field->typedPointer));
      break;
    case AQDB_DataType_Timestamp:
      GWEN_Time_free((GWEN_TIME*)(field->typedPointer));
      break;
    case AQDB_DataType_Amount:
      AQDB_Value_free((AQDB_VALUE*)(field->typedPointer));
      break;
    default:
      break;
    }
    field->typedPointer=NULL;
    field->type=AQDB_DataType_None;
  }
}



void AQDB_ObjectField_Attach(AQDB_OBJECT_FIELD *field) {
  assert(field);
  assert(field->refCount);
  field->refCount++;
}



#if 0
AQDB_OBJECT_FIELD *AQDB_ObjectField_dup(const AQDB_OBJECT_FIELD *field) {
  AQDB_OBJECT_FIELD *nf;

  assert(field);
  assert(field->refCount);

  nf=AQDB_ObjectField_new(field->dataBegin, field->fieldLen);
  if (field->typedPointer!=NULL) {
    nf->type=field->type;

    switch(field->type) {
    case AQDB_DataType_Date:
      nf->typedPointer=GWEN_Date_dup((GWEN_DATE*)(field->typedPointer));
      break;
    case AQDB_DataType_Timestamp:
      nf->typedPointer=GWEN_Time_dup((GWEN_TIME*)(field->typedPointer));
      break;
    case AQDB_DataType_Amount:
      nf->typedPointer=AQDB_Value_dup((AQDB_VALUE*)(field->typedPointer));
      break;

    default:
      break;
    }
  }

  return nf;
}
#endif



AQDB_OBJECT_SHARED *AQDB_ObjectShared_new(AQDB_ID tableId,
					  AQDB_ID objectId,
					  uint32_t nfields) {
  AQDB_OBJECT_SHARED *so;
  uint32_t flen;

  assert(nfields>0);
  flen=nfields*sizeof(AQDB_OBJECT_FIELD*);

  so=(AQDB_OBJECT_SHARED*) GWEN_Memory_malloc(sizeof(AQDB_OBJECT_SHARED)+flen);
  assert(so);
  memset(so, 0, sizeof(AQDB_OBJECT_SHARED)+flen);

  so->refCount=1;
  so->tableId=tableId;
  so->objectId=objectId;
  so->fieldCount=nfields;

  return so;
}



void AQDB_ObjectShared_Attach(AQDB_OBJECT_SHARED *so) {
  assert(so);
  assert(so->refCount);
  so->refCount++;
}




void AQDB_ObjectShared_free(AQDB_OBJECT_SHARED *so) {
  if (so) {
    assert(so->refCount);
    if (so->refCount==1) {
      int i;

      /* clear field translations and fields */
      for (i=0; i<so->fieldCount; i++) {
	if (so->fieldPointers[i])
	  AQDB_ObjectField_free(so->fieldPointers[i]);
	so->fieldPointers[i]=NULL;
      }

      GWEN_Memory_dealloc((void*) so);
    }
    else
      so->refCount--;
  }
}



AQDB_OBJECT_SHARED *AQDB_ObjectShared_dup(const AQDB_OBJECT_SHARED *so) {
  AQDB_OBJECT_SHARED *nso;
  int i;

  assert(so);
  assert(so->refCount);

  nso=AQDB_ObjectShared_new(so->tableId, so->objectId, so->fieldCount);

  /* copy field translations and fields */
  for (i=0; i<so->fieldCount; i++) {
    if (so->fieldPointers[i]) {
      AQDB_ObjectField_Attach(so->fieldPointers[i]);
      nso->fieldPointers[i]=so->fieldPointers[i];
    }
  }

  return nso;
}









AQDB_OBJECT *AQDB_Object_new(AQDB_ID tableId,
			     AQDB_ID objectId,
			     uint32_t nfields) {
  AQDB_OBJECT *o;

  GWEN_NEW_OBJECT(AQDB_OBJECT, o);
  o->refCount=1;
  GWEN_LIST_INIT(AQDB_OBJECT, o);

  o->shared=AQDB_ObjectShared_new(tableId, objectId, nfields);

  return o;
}



AQDB_OBJECT *AQDB_Object_dup(const AQDB_OBJECT *o) {
  AQDB_OBJECT *no;

  GWEN_NEW_OBJECT(AQDB_OBJECT, no);
  no->refCount=1;
  GWEN_LIST_INIT(AQDB_OBJECT, no);

  AQDB_ObjectShared_Attach(o->shared);
  no->shared=o->shared;

  return no;
}



void AQDB_Object_Attach(AQDB_OBJECT *o) {
  assert(o);
  assert(o->refCount);
  o->refCount++;
}




void AQDB_Object_free(AQDB_OBJECT *o) {
  if (o) {
    assert(o->refCount);
    if ((--o->refCount)==0) {
      GWEN_LIST_FINI(AQDB_OBJECT, o);
      AQDB_ObjectShared_free(o->shared);
      GWEN_FREE_OBJECT(o);
    }
  }
}




int AQDB_Object_GetField(const AQDB_OBJECT *o,
			 uint32_t idx,
			 uint8_t **pData,
			 uint32_t *pLen) {
  assert(o);
  assert(o->refCount);

  if (idx>=o->shared->fieldCount) {
    DBG_INFO(AQDB_LOGDOMAIN, "Field index out of range (%d>%d)",
	     idx, o->shared->fieldCount);
    *pData=NULL;
    *pLen=0;
    return 0;
  }
  else {
    if (o->shared->fieldPointers[idx]) {
      *pData=o->shared->fieldPointers[idx]->dataBegin;
      *pLen=o->shared->fieldPointers[idx]->fieldLen;
    }
    else {
      *pData=NULL;
      *pLen=0;
    }
    return 0;
  }
}



int AQDB_Object_SetField(AQDB_OBJECT *o,
			 uint32_t idx,
			 const uint8_t *data,
			 uint32_t len) {
  assert(o);
  assert(o->refCount);

  if (idx>=o->shared->fieldCount) {
    DBG_INFO(AQDB_LOGDOMAIN, "Field index out of range (%d>%d)",
	     idx, o->shared->fieldCount);
    return GWEN_ERROR_INVALID;
  }
  else {
    /* create a private copy of the shared data */
    if (o->shared->refCount>1) {
      AQDB_OBJECT_SHARED *nso;

      nso=AQDB_ObjectShared_dup(o->shared);
      AQDB_ObjectShared_free(o->shared);
      o->shared=nso;
    }

    if (o->shared->fieldPointers[idx])
      AQDB_ObjectField_free(o->shared->fieldPointers[idx]);
    o->shared->fieldPointers[idx]=AQDB_ObjectField_new(data, len);
    return 0;
  }
}




AQDB_ID AQDB_Object_GetObjectId(const AQDB_OBJECT *o) {
  assert(o);
  assert(o->refCount);
  return o->shared->objectId;
}



void AQDB_Object_SetObjectId(AQDB_OBJECT *o, AQDB_ID objectId) {
  assert(o);
  assert(o->refCount);
  o->shared->objectId=objectId;
}



AQDB_ID AQDB_Object_GetTableId(const AQDB_OBJECT *o) {
  assert(o);
  assert(o->refCount);
  return o->shared->tableId;
}



void AQDB_Object_SetTableId(AQDB_OBJECT *o, AQDB_ID objectId) {
  assert(o);
  assert(o->refCount);
  o->shared->tableId=objectId;
}



uint32_t AQDB_Object_GetFieldCount(const AQDB_OBJECT *o) {
  assert(o);
  assert(o->refCount);
  return o->shared->fieldCount;
}



int AQDB_Object_GetFieldInt64(const AQDB_OBJECT *o,
			      uint32_t idx,
			      int64_t *pVal) {
  uint8_t *d;
  uint32_t l;
  int64_t result=0;
  int rv;

  assert(o);
  assert(o->refCount);

  rv=AQDB_Object_GetField(o, idx, &d, &l);
  if (rv) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Error getting field %d (%d)", idx, rv);
    return rv;
  }

  if (l>sizeof(int64_t)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Too many bytes for type (%d)", l);
    return GWEN_ERROR_BAD_DATA;
  }

  while(l--) {
    result<<=8;
    result+=*(d++);
  }

  *pVal=result;
  return 0;
}



int AQDB_Object_GetFieldInt32(const AQDB_OBJECT *o,
			      uint32_t idx,
			      int32_t *pVal) {
  uint8_t *d;
  uint32_t l;
  int32_t result=0;
  int rv;

  assert(o);
  assert(o->refCount);

  rv=AQDB_Object_GetField(o, idx, &d, &l);
  if (rv) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Error getting field %d (%d)", idx, rv);
    return rv;
  }

  if (l>sizeof(int32_t)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Too many bytes for type (%d)", l);
    return GWEN_ERROR_BAD_DATA;
  }

  while(l--) {
    result<<=8;
    result+=*(d++);
  }

  *pVal=result;
  return 0;
}



int AQDB_Object_GetFieldDate(AQDB_OBJECT *o,
			     uint32_t idx,
			     const GWEN_DATE **pVal) {
  uint8_t *d;
  uint32_t l;
  int rv;

  assert(o);
  assert(o->refCount);

  if (idx>=o->shared->fieldCount) {
    DBG_INFO(AQDB_LOGDOMAIN, "Field index out of range (%d>%d)",
	     idx, o->shared->fieldCount);
    return GWEN_ERROR_INVALID;
  }

  /* check whether this field is already translated to this wanted type */
  if (o->shared->fieldPointers[idx]) {
    if (o->shared->fieldPointers[idx]->type!=AQDB_DataType_Date) {
      /* it isn't, clear field and set */
      AQDB_ObjectField_ClearTypedPointer(o->shared->fieldPointers[idx]);

      rv=AQDB_Object_GetField(o, idx, &d, &l);
      if (rv<0) {
	DBG_ERROR(AQDB_LOGDOMAIN, "Error getting field %d (%d)", idx, rv);
	return rv;
      }
    
      if (l) {
	GWEN_DATE *ti;
    
	ti=GWEN_Date_fromString((const char*) d);
	if (ti) {
	  o->shared->fieldPointers[idx]->type=AQDB_DataType_Date;
	  o->shared->fieldPointers[idx]->typedPointer=(void*) ti;
	}
      }
    }

    *pVal=(const GWEN_DATE*)(o->shared->fieldPointers[idx]->typedPointer);
  }
  else
    *pVal=NULL;

  return 0;
}



int AQDB_Object_SetFieldDate(AQDB_OBJECT *o,
                             uint32_t idx,
                             const GWEN_DATE *val) {
  int rv;

  assert(o);
  assert(o->refCount);

  if (idx>=o->shared->fieldCount) {
    DBG_INFO(AQDB_LOGDOMAIN, "Field index out of range (%d>%d)",
	     idx, o->shared->fieldCount);
    return GWEN_ERROR_INVALID;
  }

  if (val) {
    rv=AQDB_Object_SetFieldString(o, idx, GWEN_Date_GetString(val));
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }
  else {
    rv=AQDB_Object_SetField(o, idx, NULL, 0);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }

  return 0;
}



int AQDB_Object_GetFieldTimestamp(AQDB_OBJECT *o,
				  uint32_t idx,
				  const GWEN_TIME **pVal) {
  uint8_t *d;
  uint32_t l;
  int rv;

  assert(o);
  assert(o->refCount);

  if (idx>=o->shared->fieldCount) {
    DBG_INFO(AQDB_LOGDOMAIN, "Field index out of range (%d>%d)",
	     idx, o->shared->fieldCount);
    return GWEN_ERROR_INVALID;
  }

  /* check whether this field is already translated to this wanted type */
  if (o->shared->fieldPointers[idx]) {
    if (o->shared->fieldPointers[idx]->type!=AQDB_DataType_Timestamp) {
      /* it isn't, clear field and set */
      AQDB_ObjectField_ClearTypedPointer(o->shared->fieldPointers[idx]);

      rv=AQDB_Object_GetField(o, idx, &d, &l);
      if (rv<0) {
	DBG_ERROR(AQDB_LOGDOMAIN, "Error getting field %d (%d)", idx, rv);
	return rv;
      }
    
      if (l) {
	GWEN_TIME *ti;

	ti=GWEN_Time_fromString((const char*) d, "YYYYMMDD-hhmmss");
	if (ti) {
	  o->shared->fieldPointers[idx]->type=AQDB_DataType_Timestamp;
	  o->shared->fieldPointers[idx]->typedPointer=(void*) ti;
	}
      }
    }

    *pVal=(const GWEN_TIME*)(o->shared->fieldPointers[idx]->typedPointer);
  }
  else
    *pVal=NULL;

  return 0;
}



int AQDB_Object_SetFieldTimestamp(AQDB_OBJECT *o,
				  uint32_t idx,
				  const GWEN_TIME *val) {
  int rv;

  assert(o);
  assert(o->refCount);

  if (idx>=o->shared->fieldCount) {
    DBG_INFO(AQDB_LOGDOMAIN, "Field index out of range (%d>%d)",
	     idx, o->shared->fieldCount);
    return GWEN_ERROR_INVALID;
  }

  if (val) {
    GWEN_BUFFER *tbuf;

    tbuf=GWEN_Buffer_new(0, 16, 0, 1);
    rv=GWEN_Time_toString(val, "YYYYMMDD-hhmmss", tbuf);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      GWEN_Buffer_free(tbuf);
      return rv;
    }

    rv=AQDB_Object_SetField(o, idx,
			    (const uint8_t*) GWEN_Buffer_GetStart(tbuf),
                            GWEN_Buffer_GetUsedBytes(tbuf)+1);
    GWEN_Buffer_free(tbuf);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }
  else {
    rv=AQDB_Object_SetField(o, idx, NULL, 0);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }
  }

  return 0;
}



int AQDB_Object_GetFieldValue(AQDB_OBJECT *o,
			      uint32_t idx,
			      const AQDB_VALUE **pVal) {
  uint8_t *d;
  uint32_t l;
  int rv;

  assert(o);
  assert(o->refCount);

  if (idx>=o->shared->fieldCount) {
    DBG_INFO(AQDB_LOGDOMAIN, "Field index out of range (%d>%d)",
	     idx, o->shared->fieldCount);
    return GWEN_ERROR_INVALID;
  }

  /* check whether this field is already translated to this wanted type */
  if (o->shared->fieldPointers[idx]) {
    if (o->shared->fieldPointers[idx]->type!=AQDB_DataType_Amount) {
      /* it isn't, clear field and set */
      AQDB_ObjectField_ClearTypedPointer(o->shared->fieldPointers[idx]);

      rv=AQDB_Object_GetField(o, idx, &d, &l);
      if (rv<0) {
	DBG_ERROR(AQDB_LOGDOMAIN, "Error getting field %d (%d)", idx, rv);
	return rv;
      }
    
      if (l) {
	AQDB_VALUE *v;

	v=AQDB_Value_fromString((const char*) d);
	if (v) {
	  o->shared->fieldPointers[idx]->type=AQDB_DataType_Amount;
	  o->shared->fieldPointers[idx]->typedPointer=(void*) v;
	}
      }
    }

    *pVal=(const AQDB_VALUE*)(o->shared->fieldPointers[idx]->typedPointer);
  }
  else
    *pVal=NULL;

  return 0;
}



int AQDB_Object_SetFieldValue(AQDB_OBJECT *o,
			      uint32_t idx,
			      const AQDB_VALUE *val) {
  int rv;

  /* TODO: this should be improved to makeuse of the GWEN_BUFFER unnecessary */
  if (val) {
    GWEN_BUFFER *tbuf;

    tbuf=GWEN_Buffer_new(0, 64, 0, 1);
    AQDB_Value_toString(val, tbuf);
    rv=AQDB_Object_SetFieldString(o, idx, GWEN_Buffer_GetStart(tbuf));
    GWEN_Buffer_free(tbuf);
  }
  else
    rv=AQDB_Object_SetFieldString(o, idx, NULL);

  return rv;
}




int AQDB_Object_SetFieldInt64(AQDB_OBJECT *o,
			      uint32_t idx,
			      int64_t val) {
  uint8_t buf[8];
  uint8_t *p;
  int rv;
  int i;

  assert(o);
  assert(o->refCount);

  buf[7]=val & 0xff;
  val>>=8;
  buf[6]=val & 0xff;
  val>>=8;
  buf[5]=val & 0xff;
  val>>=8;
  buf[4]=val & 0xff;
  val>>=8;
  buf[3]=val & 0xff;
  val>>=8;
  buf[2]=val & 0xff;
  val>>=8;
  buf[1]=val & 0xff;
  val>>=8;
  buf[0]=val & 0xff;

  p=buf;
  for (i=0; i<7; i++) {
    if (*p!=0)
      break;
    p++;
  }

  rv=AQDB_Object_SetField(o, idx, p, 8-i);
  if (rv) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Error setting field %d (%d)", idx, rv);
    return rv;
  }
  return 0;
}



int AQDB_Object_SetFieldInt32(AQDB_OBJECT *o,
			      uint32_t idx,
			      int32_t val) {
  uint8_t buf[4];
  uint8_t *p;
  int rv;
  int i;

  assert(o);
  assert(o->refCount);

  buf[3]=val & 0xff;
  val>>=8;
  buf[2]=val & 0xff;
  val>>=8;
  buf[1]=val & 0xff;
  val>>=8;
  buf[0]=val & 0xff;

  p=buf;
  for (i=0; i<3; i++) {
    if (*p!=0)
      break;
    p++;
  }

  rv=AQDB_Object_SetField(o, idx, p, 4-i);
  if (rv) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Error setting field %d (%d)", idx, rv);
    return rv;
  }
  return 0;
}



int AQDB_Object_SetFieldString(AQDB_OBJECT *o,
			       uint32_t idx,
			       const char *s) {
  int rv;

  assert(o);
  assert(o->refCount);

  if (!s)
    s="";
  rv=AQDB_Object_SetField(o, idx, (const uint8_t*)s, strlen(s)+1);
  if (rv) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Error setting field %d (%d)", idx, rv);
    return rv;
  }
  return 0;
}



int AQDB_Object_GetFieldString(const AQDB_OBJECT *o,
			       uint32_t idx,
			       const char **pVal) {
  uint8_t *d;
  uint32_t l;
  int rv;

  assert(o);
  assert(o->refCount);

  rv=AQDB_Object_GetField(o, idx, &d, &l);
  if (rv) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Error getting field %d (%d)", idx, rv);
    return rv;
  }

  if (l)
    *pVal=(const char*)d;
  else
    *pVal="";
  return 0;
}



void AQDB_Object_Dump(const AQDB_OBJECT *o, int indent) {
  int i;
  int j;

  for (i=0; i<indent; i++) fprintf(stderr, " ");
  fprintf(stderr, "Object:\n");
  for (i=0; i<indent+2; i++) fprintf(stderr, " ");
  fprintf(stderr, "Table Id  : %08x\n", o->shared->tableId);
  for (i=0; i<indent+2; i++) fprintf(stderr, " ");
  fprintf(stderr, "Object Id : %08x\n", o->shared->objectId);
  for (i=0; i<indent+2; i++) fprintf(stderr, " ");
  fprintf(stderr, "Field num : %d\n", o->shared->fieldCount);

  for (i=0; i<indent+2; i++) fprintf(stderr, " ");
  fprintf(stderr, "Fields\n");
  for (j=0; j<o->shared->fieldCount; j++) {
    uint8_t *data;
    uint32_t len;
    int rv;

    rv=AQDB_Object_GetField(o, j, &data, &len);
    if (rv<0) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Error getting field %d: %d", j, rv);
    }
    else {
      for (i=0; i<indent+4; i++) fprintf(stderr, " ");
      fprintf(stderr, "Field %02d:\n", j);
      GWEN_Text_DumpString((const char*) data, len, indent+6);
    }
  }
}



