/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqFinance (c) by 2015 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#define DISABLE_DEBUGLOG


#include "dbfile_p.h"
#include <aqdatabase/aqdb_db_be.h>
#include <aqdatabase/aqdb_object_be.h>
#include <aqdatabase/aqdb_sessionlog.h>
#include <aqdatabase/aqdb_sessiondescr.h>
#include <aqdatabase/aqdb_query.h>
#include <aqdatabase/aqdb_objectquery.h>

#include <gwenhywfar/misc.h>
#include <gwenhywfar/debug.h>
#include <gwenhywfar/directory.h>
#include <gwenhywfar/text.h>

#include <errno.h>
#include <string.h>

#include <endian.h>



GWEN_INHERIT(AQDB_DB, AQDB_DB_FILE)

#define AQDB_DBFILE_LOCK_TIMEOUT 30000



#include "dbfile_object.c"
#include "dbfile_table.c"
#include "dbfile_init.c"
#include "dbfile_query.c"
#include "dbfile_rollback.c"





AQDB_DB *AQDB_DbFile_new(const char *urlString) {
  AQDB_DB *db;
  AQDB_DB_FILE *xdb;
  const GWEN_URL *url;

  db=AQDB_DB_new(urlString);
  if (db==NULL)
    return NULL;

  GWEN_NEW_OBJECT(AQDB_DB_FILE, xdb);
  if (xdb==NULL) {
    AQDB_DB_free(db);
    return NULL;
  }
  GWEN_INHERIT_SETDATA(AQDB_DB, AQDB_DB_FILE, db, xdb,
		       AQDB_DbFile_FreeData);

  url=AQDB_DB_GetUrl(db);
  if (url) {
    GWEN_BUFFER *pbuf;
    const char *s;

    pbuf=GWEN_Buffer_new(0, 256, 0, 1);
    s=GWEN_Url_GetServer(url);
    if (s)
      GWEN_Buffer_AppendString(pbuf, s);
    s=GWEN_Url_GetPath(url);
    if (s) {
      if (GWEN_Buffer_GetUsedBytes(pbuf))
	GWEN_Buffer_AppendString(pbuf, GWEN_DIR_SEPARATOR_S);
      GWEN_Buffer_AppendString(pbuf, s);
    }
    xdb->dataFileName=strdup(GWEN_Buffer_GetStart(pbuf));
    GWEN_Buffer_free(pbuf);
  }

  xdb->queryResultList=AQDB_QueryResult_List_new();

  /* set virtual functions */
  AQDB_DB_SetCreateFn(db, AQDB_DbFile_Create);
  AQDB_DB_SetOpenFn(db, AQDB_DbFile_Open);
  AQDB_DB_SetCloseFn(db, AQDB_DbFile_Close);

  AQDB_DB_SetCreateTableFn(db, AQDB_DbFile_CreateTable);
  AQDB_DB_SetOpenTableFn(db, AQDB_DbFile_OpenTable);
  AQDB_DB_SetCloseTableFn(db, AQDB_DbFile_CloseTable);

  AQDB_DB_SetAddObjectFn(db, AQDB_DbFile_AddObject);
  AQDB_DB_SetDeleteObjectFn(db, AQDB_DbFile_DeleteObject);
  AQDB_DB_SetWriteObjectFn(db, AQDB_DbFile_WriteObject);
  AQDB_DB_SetReadObjectFn(db, AQDB_DbFile_ReadObject);

  AQDB_DB_SetBeginEditFn(db, AQDB_DbFile_BeginEdit);
  AQDB_DB_SetEndEditFn(db, AQDB_DbFile_EndEdit);

  AQDB_DB_SetGetColumnListFn(db, AQDB_DbFile_GetColumnList);

  AQDB_DB_SetQuerySubmitFn(db, AQDB_DbFile_QuerySubmit);
  AQDB_DB_SetQueryGetFirstFn(db, AQDB_DbFile_QueryGetFirst);
  AQDB_DB_SetQueryGetNextFn(db, AQDB_DbFile_QueryGetNext);
  AQDB_DB_SetQueryCloseFn(db, AQDB_DbFile_QueryClose);

  AQDB_DB_SetGetRollbackSessionsFn(db, AQDB_DbFile_GetRollbackSessions);
  AQDB_DB_SetUndoSessionFn(db, AQDB_DbFile_UndoSession);

  return db;
}



void GWENHYWFAR_CB AQDB_DbFile_FreeData(void *bp, void *p) {
  AQDB_DB_FILE *xdb;

  xdb=(AQDB_DB_FILE*)p;
  assert(xdb);

  if (xdb->fileLock) {
    DBG_WARN(AQDB_LOGDOMAIN, "Database still locked, unlocking");
    GWEN_FSLock_Unlock(xdb->fileLock);
    GWEN_FSLock_free(xdb->fileLock);
  }

  AQDB_QueryResult_List_free(xdb->queryResultList);

  free(xdb->dataFileName);
  free(xdb->currentUserName);

  GWEN_FREE_OBJECT(xdb);
}



const char *AQDB_DbFile_GetDataFileName(const AQDB_DB *db) {
  AQDB_DB_FILE *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  return xdb->dataFileName;
}



void AQDB_DbFile_SetCurrentUserName(AQDB_DB *db, const char *userName) {
  AQDB_DB_FILE *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  free(xdb->currentUserName);
  if (userName)
    xdb->currentUserName=strdup(userName);
  else
    xdb->currentUserName=NULL;
}



const char *AQDB_DbFile_GetCurrentUserName(const AQDB_DB *db) {
  AQDB_DB_FILE *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  return xdb->currentUserName;
}



int GWENHYWFAR_CB AQDB_DbFile_Create(AQDB_DB *db, uint32_t flags) {
  AQDB_DB_FILE *xdb;
  AQFS_FILE *fsf;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  DBG_INFO(AQDB_LOGDOMAIN, "Creating database");

  /* create file */
  fsf=AQFS_File_new(xdb->dataFileName);
  rv=AQFS_File_Create(fsf);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQFS_File_free(fsf);
    return rv;
  }
  xdb->file=fsf;

  rv=AQDB_DbFile_CreateDatabaseStart(db);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQFS_File_free(xdb->file);
    xdb->file=NULL;
    return rv;
  }

  DBG_INFO(AQDB_LOGDOMAIN, "Closing file");
  rv=AQFS_File_Close(xdb->file);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d), ignoring", rv);
    AQDB_DbFile_ReleaseInternalStructures(db);
    return rv;
  }

  AQDB_DbFile_ReleaseInternalStructures(db);
  xdb->openFlags=0;


  DBG_INFO(AQDB_LOGDOMAIN, "Database created");
  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_Open(AQDB_DB *db, uint32_t flags) {
  AQDB_DB_FILE *xdb;
  AQFS_FILE *fsf;
  int rv;
  AQDB_ID tid;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  DBG_INFO(AQDB_LOGDOMAIN, "Opening database");

  if ((flags & (AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE))==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "No open flags given (%d)", flags);
    return GWEN_ERROR_INVALID;
  }

  /* open file */
  fsf=AQFS_File_new(xdb->dataFileName);
  rv=AQFS_File_Open(fsf, (flags & AQDB_ACTION_FLAGS_WRITE)?1:0);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQFS_File_free(fsf);
    return rv;
  }
  xdb->file=fsf;

  /* read database headers and stuff */
  rv=AQDB_DbFile_ReadDatabaseStart(db);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DbFile_ReleaseInternalStructures(db);
    return rv;
  }

  /* read types */
  rv=AQDB_DbFile_ReadTypes(db);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DbFile_ReleaseInternalStructures(db);
    return rv;
  }

  /* open session description type */
  tid=0;
  rv=AQDB_DbFile_ReallyOpenTable(db,
                                 0,
                                 AQDB_DBFILE_TYPENAME_SESSIONS,
                                 flags & AQDB_ACTION_FLAGS_MASK,
                                 &tid);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DbFile_ReleaseInternalStructures(db);
    return rv;
  }
  xdb->tableIdSessionDescr=tid;

  /* open changelog type */
  tid=0;
  rv=AQDB_DbFile_ReallyOpenTable(db,
                                 0,
                                 AQDB_DBFILE_TYPENAME_CHANGELOG,
                                 flags & AQDB_ACTION_FLAGS_MASK,
                                 &tid);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DbFile_ReleaseInternalStructures(db);
    return rv;
  }
  xdb->tableIdChangeLog=tid;


  /* database is open */
  xdb->openFlags=flags;

  DBG_INFO(AQDB_LOGDOMAIN, "Database is now open");
  return rv;
}



int GWENHYWFAR_CB AQDB_DbFile_Close(AQDB_DB *db, uint32_t flags) {
  AQDB_DB_FILE *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  DBG_INFO(AQDB_LOGDOMAIN, "Closing database");

  if (xdb->openFlags==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "DB not open");
    return GWEN_ERROR_NOT_OPEN;
  }
  else {
    /* TODO: close all open tables */
  }

  /* close file */
  if (xdb->file) {
    DBG_INFO(AQDB_LOGDOMAIN, "Closing file");
    rv=AQFS_File_Close(xdb->file);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d), ignoring", rv);
    }
  }

  AQDB_DbFile_ReleaseInternalStructures(db);
  xdb->openFlags=0;

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_CreateTable(AQDB_DB *db,
                                          const char *name,
                                          uint32_t flags,
                                          const AQDB_COLUMN_LIST *columns) {
  if (name && *name!='+')
    return AQDB_DbFile_ReallyCreateTable(db, name, flags, columns);
  else {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database type must not start with a plus sign");
    return GWEN_ERROR_INVALID;
  }
}



int GWENHYWFAR_CB AQDB_DbFile_OpenTable(AQDB_DB *db,
                                        uint32_t aflags,
                                        const char *name,
                                        uint32_t flags,
                                        AQDB_ID *pTableId) {
  if (name && *name!='+')
    return AQDB_DbFile_ReallyOpenTable(db, aflags, name, flags, pTableId);
  else {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database type must not start with a plus sign");
    return GWEN_ERROR_INVALID;
  }
}



int GWENHYWFAR_CB AQDB_DbFile_CloseTable(AQDB_DB *db,
                                         AQDB_ID tableId,
                                         uint32_t flags) {
  AQDB_DB_FILE *xdb;
  AQDB_DBFILE_TABLE *dbt;
  uint32_t oflags;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  dbt=AQDB_DbFile_GetTableById(db, tableId);
  if (dbt==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table with id %d not found.", tableId);
    return GWEN_ERROR_INVALID;
  }

  oflags=AQDB_DbFile_Table_GetOpenFlags(dbt);
  if ((oflags & AQDB_ACTION_FLAGS_MASK)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table %d not open.", tableId);
    return GWEN_ERROR_INVALID;
  }

  return AQDB_DbFile_ReallyCloseTable(db, tableId, flags);
}



int GWENHYWFAR_CB AQDB_DbFile_AddObject(AQDB_DB *db,
                                        AQDB_ID tableId,
                                        AQDB_OBJECT *object) {
  AQDB_DB_FILE *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_WRITE)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for writing");
    return GWEN_ERROR_PERMISSIONS;
  }

  if (!(xdb->fileLock)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not in edit mode, you must call AQDB_DB_BeginEdit before.");
    return GWEN_ERROR_WRITE;
  }

  if (1) {
    AQDB_DBFILE_TABLE *dbt;
    AQDB_DBFILE_CHANGELOG *chg;
    AQFS_TABLE_LIST *ftl;
    int rv;
    int64_t rd;
    uint64_t offset;
    AQDB_OBJECT *o;
    AQDB_ID objectId;

    /* get type table */
    dbt=AQDB_DbFile_GetTableById(db, tableId);
    if (dbt==NULL) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Table %d not found", (int) tableId);
      return GWEN_ERROR_NOT_FOUND;
    }
    if (!(AQDB_DbFile_Table_GetOpenFlags(dbt) & AQDB_ACTION_FLAGS_WRITE)) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Table is not open for writing");
      return GWEN_ERROR_PERMISSIONS;
    }

    ftl=AQDB_DbFile_Table_GetTableListObjects(dbt);

    /* prepare entry for changelog event */
    chg=AQDB_DbFile_Changelog_new();
    AQDB_DbFile_Changelog_SetSessionLogId(chg, xdb->currentSessionId);
    AQDB_DbFile_Changelog_SetTableId(chg, tableId);
    AQDB_DbFile_Changelog_SetOldOffset(chg, 0);
    AQDB_DbFile_Changelog_SetEntryType(chg, AQDB_DbFile_ChangelogType_Addition);

    /* really add object to database */
    rv=AQDB_DbFile_ReallyAddObject(db, tableId, object);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_DbFile_Changelog_free(chg);
      return rv;
    }

    /* object id is assigned by AQDB_DbFile_ReallyAddObject() */
    objectId=AQDB_Object_GetObjectId(object);
    if (objectId==0) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid object id after adding, SNH!");
      AQDB_DbFile_Changelog_free(chg);
      return GWEN_ERROR_INTERNAL;
    }
    AQDB_DbFile_Changelog_SetObjectId(chg, objectId);

    /* find new object position within the table */
    rd=AQFS_File_GetTableEntry(xdb->file, ftl, objectId-1);
    if (rd<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
      return (int) rd;
    }
    offset=rd;
    AQDB_DbFile_Changelog_SetNewOffset(chg, offset);

    /* transform changelog entry to database object */
    o=NULL;
    rv=AQDB_DbFile_Changelog_CreateObject(db, xdb->tableIdChangeLog, 0, &o);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_DbFile_Changelog_free(chg);
      return rv;
    }

    rv=AQDB_DbFile_Changelog_WriteObject(chg, o);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_Object_free(o);
      AQDB_DbFile_Changelog_free(chg);
      return rv;
    }

    /* append change log entry to database */
    rv=AQDB_DbFile_ReallyAddObject(db, xdb->tableIdChangeLog, o);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_Object_free(o);
      AQDB_DbFile_Changelog_free(chg);
      return rv;
    }
    AQDB_Object_free(o);
    AQDB_DbFile_Changelog_free(chg);

    /* increase number of additions */
    if (xdb->currentSession)
      AQDB_SessionDescr_IncAdditions(xdb->currentSession);
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_DeleteObject(AQDB_DB *db, AQDB_ID tableId, AQDB_ID objectId) {
  AQDB_DB_FILE *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_WRITE)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for writing");
    return GWEN_ERROR_PERMISSIONS;
  }

  if (!(xdb->fileLock)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not in edit mode, you must call AQDB_DB_BeginEdit before.");
    return GWEN_ERROR_WRITE;
  }

  if (objectId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid object id.");
    return GWEN_ERROR_INVALID;
  }

  if (tableId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id.");
    return GWEN_ERROR_INVALID;
  }

  if (1) {
    AQDB_DBFILE_TABLE *dbt;
    AQDB_DBFILE_CHANGELOG *chg;
    AQFS_TABLE_LIST *ftl;
    int rv;
    int64_t rd;
    uint64_t offset;
    AQDB_OBJECT *o;

    /* get type table */
    dbt=AQDB_DbFile_GetTableById(db, tableId);
    if (dbt==NULL) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Table %d not found", (int) tableId);
      return GWEN_ERROR_NOT_FOUND;
    }
    if (!(AQDB_DbFile_Table_GetOpenFlags(dbt) & AQDB_ACTION_FLAGS_WRITE)) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Table is not open for writing");
      return GWEN_ERROR_PERMISSIONS;
    }

    ftl=AQDB_DbFile_Table_GetTableListObjects(dbt);

    /* find current object position within the table */
    rd=AQFS_File_GetTableEntry(xdb->file, ftl, objectId-1);
    if (rd<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
      return (int) rd;
    }
    offset=rd;

    /* prepare entry for changelog event */
    chg=AQDB_DbFile_Changelog_new();
    AQDB_DbFile_Changelog_SetSessionLogId(chg, xdb->currentSessionId);
    AQDB_DbFile_Changelog_SetTableId(chg, tableId);
    AQDB_DbFile_Changelog_SetObjectId(chg, objectId);
    AQDB_DbFile_Changelog_SetOldOffset(chg, offset);
    AQDB_DbFile_Changelog_SetNewOffset(chg, 0);
    AQDB_DbFile_Changelog_SetEntryType(chg, AQDB_DbFile_ChangelogType_Modification);

    /* set new object position to 0 */
    rd=AQFS_File_SetTableEntry(xdb->file, ftl, objectId-1, 0);
    if (rd<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
      return (int) rd;
    }

    /* transform changelog entry to database object */
    o=NULL;
    rv=AQDB_DbFile_Changelog_CreateObject(db, xdb->tableIdChangeLog, 0, &o);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_DbFile_Changelog_free(chg);
      return rv;
    }

    rv=AQDB_DbFile_Changelog_WriteObject(chg, o);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_Object_free(o);
      AQDB_DbFile_Changelog_free(chg);
      return rv;
    }

    /* append change log entry to database */
    rv=AQDB_DbFile_ReallyAddObject(db, xdb->tableIdChangeLog, o);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_Object_free(o);
      AQDB_DbFile_Changelog_free(chg);
      return rv;
    }
    AQDB_Object_free(o);
    AQDB_DbFile_Changelog_free(chg);

    /* increase number of deletions */
    if (xdb->currentSession)
      AQDB_SessionDescr_IncDeletions(xdb->currentSession);
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_WriteObject(AQDB_DB *db, const AQDB_OBJECT *object) {
  AQDB_DB_FILE *xdb;
  AQDB_ID tableId;
  AQDB_ID objectId;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_WRITE)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for writing");
    return GWEN_ERROR_PERMISSIONS;
  }

  if (!(xdb->fileLock)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not in edit mode, you must call AQDB_DB_BeginEdit before.");
    return GWEN_ERROR_WRITE;
  }

  objectId=AQDB_Object_GetObjectId(object);
  if (objectId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid object id.");
    return GWEN_ERROR_INVALID;
  }

  tableId=AQDB_Object_GetTableId(object);
  if (tableId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id.");
    return GWEN_ERROR_INVALID;
  }

  if (1) {
    AQDB_DBFILE_TABLE *dbt;
    AQDB_DBFILE_CHANGELOG *chg;
    AQFS_TABLE_LIST *ftl;
    int rv;
    int64_t rd;
    uint64_t offset;
    AQDB_OBJECT *o;

    /* get type table */
    dbt=AQDB_DbFile_GetTableById(db, tableId);
    if (dbt==NULL) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Table %d not found", (int) tableId);
      return GWEN_ERROR_NOT_FOUND;
    }
    if (!(AQDB_DbFile_Table_GetOpenFlags(dbt) & AQDB_ACTION_FLAGS_WRITE)) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Table is not open for writing");
      return GWEN_ERROR_PERMISSIONS;
    }

    ftl=AQDB_DbFile_Table_GetTableListObjects(dbt);

    /* find current object position within the table */
    rd=AQFS_File_GetTableEntry(xdb->file, ftl, objectId-1);
    if (rd<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
      return (int) rd;
    }
    offset=rd;

    /* prepare entry for changelog event */
    chg=AQDB_DbFile_Changelog_new();
    AQDB_DbFile_Changelog_SetSessionLogId(chg, xdb->currentSessionId);
    AQDB_DbFile_Changelog_SetTableId(chg, tableId);
    AQDB_DbFile_Changelog_SetObjectId(chg, objectId);
    AQDB_DbFile_Changelog_SetOldOffset(chg, offset);
    AQDB_DbFile_Changelog_SetEntryType(chg, AQDB_DbFile_ChangelogType_Modification);

    /* really write object to database */
    rv=AQDB_DbFile_ReallyWriteObject(db, object);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_DbFile_Changelog_free(chg);
      return rv;
    }

    /* find new object position within the table */
    rd=AQFS_File_GetTableEntry(xdb->file, ftl, objectId-1);
    if (rd<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
      return (int) rd;
    }
    offset=rd;
    AQDB_DbFile_Changelog_SetNewOffset(chg, offset);

    /* transform changelog entry to database object */
    o=NULL;
    rv=AQDB_DbFile_Changelog_CreateObject(db, xdb->tableIdChangeLog, 0, &o);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_DbFile_Changelog_free(chg);
      return rv;
    }

    rv=AQDB_DbFile_Changelog_WriteObject(chg, o);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_Object_free(o);
      AQDB_DbFile_Changelog_free(chg);
      return rv;
    }

    /* append change log entry to database */
    rv=AQDB_DbFile_ReallyAddObject(db, xdb->tableIdChangeLog, o);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_Object_free(o);
      AQDB_DbFile_Changelog_free(chg);
      return rv;
    }
    AQDB_Object_free(o);
    AQDB_DbFile_Changelog_free(chg);

    /* increase number of modifications */
    if (xdb->currentSession)
      AQDB_SessionDescr_IncModifications(xdb->currentSession);
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_ReadObject(AQDB_DB *db,
                                         uint32_t aflags,
                                         AQDB_ID tableId,
                                         AQDB_ID objectId,
                                         AQDB_OBJECT **pObject) {
  AQDB_DB_FILE *xdb;
  AQDB_DBFILE_TABLE *dbt;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (objectId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid object id 0");
    return GWEN_ERROR_INVALID;
  }

  if (tableId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id 0");
    return GWEN_ERROR_INVALID;
  }

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_READ)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for reading");
    return GWEN_ERROR_PERMISSIONS;
  }

  dbt=AQDB_DbFile_GetTableById(db, tableId);
  if (dbt==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table [%d] not found", (int) tableId);
    return GWEN_ERROR_NOT_FOUND;
  }

  return AQDB_DbFile_ReallyReadObject(db, aflags, tableId, objectId, pObject);
}



int GWENHYWFAR_CB AQDB_DbFile_BeginEdit(AQDB_DB *db, uint32_t editFlags, const char *user) {
  AQDB_DB_FILE *xdb;
  int rv;
  GWEN_FSLOCK *lck;
  GWEN_FSLOCK_RESULT res;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  DBG_INFO(AQDB_LOGDOMAIN, "Begin edit session");

  if ((xdb->openFlags & AQDB_ACTION_FLAGS_WRITE)==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database not open for writing");
    return GWEN_ERROR_INVALID;
  }

  if (xdb->fileLock!=NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Already locked");
    return GWEN_ERROR_INVALID;
  }

  lck=GWEN_FSLock_new(xdb->dataFileName, GWEN_FSLock_TypeFile);
  res=GWEN_FSLock_Lock(lck, AQDB_DBFILE_LOCK_TIMEOUT, 0);
  switch(res) {
    case GWEN_FSLock_ResultOk:
      xdb->fileLock=lck;
      break;
    case GWEN_FSLock_ResultError:
      DBG_INFO(AQDB_LOGDOMAIN, "Lock error");
      GWEN_FSLock_free(lck);
      return GWEN_ERROR_IO;
    case GWEN_FSLock_ResultTimeout:
      DBG_INFO(AQDB_LOGDOMAIN, "Lock timeout");
      GWEN_FSLock_free(lck);
      return GWEN_ERROR_TIMEOUT;
    case GWEN_FSLock_ResultBusy:
      DBG_INFO(AQDB_LOGDOMAIN, "Lock timeout (busy)");
      GWEN_FSLock_free(lck);
      return GWEN_ERROR_TIMEOUT;
    case GWEN_FSLock_ResultNoLock:
      DBG_INFO(AQDB_LOGDOMAIN, "Invalid lock");
      GWEN_FSLock_free(lck);
      return GWEN_ERROR_IO;
    case GWEN_FSLock_ResultUserAbort:
      DBG_INFO(AQDB_LOGDOMAIN, "User aborted");
      GWEN_FSLock_free(lck);
      return GWEN_ERROR_USER_ABORTED;
    default:
      DBG_INFO(AQDB_LOGDOMAIN, "Unknown error");
      GWEN_FSLock_free(lck);
      return GWEN_ERROR_GENERIC;
  }

  /* set username for session logs */
  AQDB_DbFile_SetCurrentUserName(db, user);
  xdb->editFlags=editFlags;

  /* get a session id */
  if (editFlags & AQDB_DBFILE_SESSION_FLAGS_ROLLBACK) {
    /* dont increment session id when in rollback mode */
    DBG_INFO(AQDB_LOGDOMAIN, "BeginEdit in rollback mode");
    xdb->currentSessionId=AQDB_DbFile_Header_GetLastSessionId(xdb->header);
  }
  else {
    xdb->currentSessionId=AQDB_DbFile_Header_GetLastSessionId(xdb->header)+1;
    AQDB_DbFile_Header_SetLastSessionId(xdb->header, xdb->currentSessionId);
  }
  rv=AQDB_DbFile_WriteHeader(db, xdb->header);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_FSLock_Unlock(xdb->fileLock);
    GWEN_FSLock_free(xdb->fileLock);
    xdb->fileLock=NULL;
    return rv;
  }

  /* let all types copy the current object list */
  if (xdb->tables) {
    AQDB_DBFILE_TABLE *dbt;

    dbt=AQDB_DbFile_Table_List_First(xdb->tables);
    while(dbt) {
      AQFS_TABLE_LIST *ftl;

      ftl=AQDB_DbFile_Table_GetTableListObjects(dbt);
      if (ftl) {
        AQFS_TABLE_LIST *ftlCopy;

        ftlCopy=AQFS_Table_List_dup(ftl);
        AQDB_DbFile_Table_SetSavedTableListObjects(dbt, ftlCopy);
      }

      dbt=AQDB_DbFile_Table_List_Next(dbt);
    }
  }

  /* create session */
  if (1) {
    AQDB_SESSION_DESCR *sess;
    GWEN_TIME *ti;

    ti=GWEN_CurrentTime();
    sess=AQDB_SessionDescr_new();
    AQDB_SessionDescr_SetSessionId(sess, xdb->currentSessionId);
    AQDB_SessionDescr_SetUserName(sess, xdb->currentUserName);
    AQDB_SessionDescr_SetDateTime(sess, ti);
    GWEN_Time_free(ti);

    if (xdb->currentSession)
      AQDB_SessionDescr_free(xdb->currentSession);
    xdb->currentSession=sess;
  }

  return 0;
}



int AQDB_DbFile_CloseSession(AQDB_DB *db) {
  AQDB_DB_FILE *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  DBG_INFO(AQDB_LOGDOMAIN, "Closing session");

  /* save session */
  if (xdb->currentSession) {
    if (AQDB_SessionDescr_GetAdditions(xdb->currentSession) ||
        AQDB_SessionDescr_GetDeletions(xdb->currentSession) ||
        AQDB_SessionDescr_GetModifications(xdb->currentSession)) {
      if (xdb->editFlags & AQDB_DBFILE_SESSION_FLAGS_ROLLBACK) {
        DBG_INFO(AQDB_LOGDOMAIN, "Rollback session, not saving session");
      }
      else {
        AQDB_DBFILE_TABLE *dbt;
    
        dbt=AQDB_DbFile_GetTableByName(db, AQDB_DBFILE_TYPENAME_SESSIONS);
        if (dbt) {
          AQDB_OBJECT *o=NULL;
    
          rv=AQDB_SessionDescr_CreateObject(db, AQDB_DbFile_Table_GetTableId(dbt), 0, &o);
          if (rv<0) {
            DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
            return rv;
          }
    
          rv=AQDB_SessionDescr_toObject(xdb->currentSession, o);
          if (rv<0) {
            DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
            AQDB_Object_free(o);
            return rv;
          }
    
          rv=AQDB_DbFile_ReallyAddObject(db, AQDB_DbFile_Table_GetTableId(dbt), o);
          if (rv<0) {
            DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
            AQDB_Object_free(o);
            return rv;
          }
          AQDB_Object_free(o);
        }
      }
    }
    else {
      DBG_INFO(AQDB_LOGDOMAIN, "No additions, deletions or modifications, not saving session");
    }
  }
  else {
    DBG_ERROR(AQDB_LOGDOMAIN, "No session descriptor table in database");
    return GWEN_ERROR_INTERNAL;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_EndEdit(AQDB_DB *db, uint32_t flags) {
  AQDB_DB_FILE *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  DBG_INFO(AQDB_LOGDOMAIN, "Ending edit session");

  if (xdb->fileLock==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Not locked, not in edit mode");
    return GWEN_ERROR_INVALID;
  }

  if (!(flags & AQDB_ACTION_FLAGS_ABORT)) {
    /* close session */
    rv=AQDB_DbFile_CloseSession(db);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      GWEN_FSLock_Unlock(xdb->fileLock);
      GWEN_FSLock_free(xdb->fileLock);
      xdb->fileLock=NULL;
      return rv;
    }
  }

  if (flags & AQDB_ACTION_FLAGS_ABORT) {
    DBG_INFO(AQDB_LOGDOMAIN, "Aborting edit session");
    /* set the saved object list in all types to effectively abort the edit session */
    if (xdb->tables) {
      AQDB_DBFILE_TABLE *dbt;
  
      dbt=AQDB_DbFile_Table_List_First(xdb->tables);
      while(dbt) {
        AQFS_TABLE_LIST *ftl;
  
        ftl=AQDB_DbFile_Table_GetSavedTableListObjects(dbt);
        if (ftl) {
          AQFS_TABLE_LIST *ftlCopy;
  
          ftlCopy=AQFS_Table_List_dup(ftl);
          AQDB_DbFile_Table_SetTableListObjects(dbt, ftlCopy);
          AQDB_DbFile_Table_SetSavedTableListObjects(dbt, NULL);
        }

        dbt=AQDB_DbFile_Table_List_Next(dbt);
      }
    }
  }
  else  {
    /* delete saved tables */
    if (xdb->tables) {
      AQDB_DBFILE_TABLE *dbt;
  
      DBG_INFO(AQDB_LOGDOMAIN, "Resetting savedTableList in types");
      dbt=AQDB_DbFile_Table_List_First(xdb->tables);
      while(dbt) {
        AQDB_DbFile_Table_SetSavedTableListObjects(dbt, NULL);
        dbt=AQDB_DbFile_Table_List_Next(dbt);
      }
    }
  }


  /* let all types write their object lists and column lists */
  rv=AQDB_DbFile_FlushTypeData(db);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_FSLock_Unlock(xdb->fileLock);
    GWEN_FSLock_free(xdb->fileLock);
    xdb->fileLock=NULL;
    return rv;
  }

  xdb->currentSessionId=0;
  GWEN_FSLock_Unlock(xdb->fileLock);
  GWEN_FSLock_free(xdb->fileLock);
  xdb->fileLock=NULL;

  return 0;
}



AQDB_COLUMN_LIST *GWENHYWFAR_CB AQDB_DbFile_GetColumnList(AQDB_DB *db,
                                                          uint32_t aflags,
                                                          AQDB_ID tableId) {
  AQDB_DB_FILE *xdb;
  AQDB_DBFILE_TABLE *dbt;
  AQDB_COLUMN_LIST *cl;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  dbt=AQDB_DbFile_GetTableById(db, tableId);
  if (dbt==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table %d not found", (int) tableId);
    return NULL;
  }

  cl=AQDB_DbFile_Table_GetColumnList(dbt);
  if (cl==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table %d has not column list, should not happen", (int) tableId);
    return NULL;
  }

  cl=AQDB_Column_List_dup(cl);

  return cl;
}



int GWENHYWFAR_CB AQDB_DbFile_QuerySubmit(AQDB_DB *db,
                                          uint32_t aflags,
                                          AQDB_ID tableId,
                                          AQDB_ID refQueryId,
                                          const char *queryExpression,
                                          uint32_t flags,
                                          AQDB_ID *pQueryId) {
  AQDB_DB_FILE *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_READ)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for reading");
    return GWEN_ERROR_PERMISSIONS;
  }

  return AQDB_DbFile_ReallyQuerySubmit(db, aflags, tableId,
                                       refQueryId, queryExpression, flags,
                                       pQueryId);
}



int GWENHYWFAR_CB AQDB_DbFile_QueryGetFirst(AQDB_DB *db,
                                            AQDB_ID queryId,
                                            AQDB_ID *pIdListPtr,
                                            int *pIdListLen) {
  AQDB_DB_FILE *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_READ)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for reading");
    return GWEN_ERROR_PERMISSIONS;
  }

  return AQDB_DbFile_ReallyQueryGetFirst(db, queryId, pIdListPtr, pIdListLen);
}



int GWENHYWFAR_CB AQDB_DbFile_QueryGetNext(AQDB_DB *db,
                                           AQDB_ID queryId,
                                           AQDB_ID *pIdListPtr,
                                           int *pIdListLen) {
  AQDB_DB_FILE *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_READ)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for reading");
    return GWEN_ERROR_PERMISSIONS;
  }

  return AQDB_DbFile_ReallyQueryGetNext(db, queryId, pIdListPtr, pIdListLen);
}



int GWENHYWFAR_CB AQDB_DbFile_QueryClose(AQDB_DB *db, AQDB_ID queryId) {
  AQDB_DB_FILE *xdb;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_READ)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for reading");
    return GWEN_ERROR_PERMISSIONS;
  }

  return AQDB_DbFile_ReallyQueryClose(db, queryId);
}



int GWENHYWFAR_CB AQDB_DbFile_GetRollbackSessions(AQDB_DB *db,
                                                  uint32_t aflags,
                                                  AQDB_SESSION_DESCR_LIST *sl) {
  AQDB_DB_FILE *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_READ)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for reading");
    return GWEN_ERROR_PERMISSIONS;
  }

  rv=AQDB_DbFile_GetSessionDescriptions(db, aflags, 0, sl);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int GWENHYWFAR_CB AQDB_DbFile_UndoSession(AQDB_DB *db, AQDB_ID sessionId) {
  AQDB_DB_FILE *xdb;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  DBG_INFO(AQDB_LOGDOMAIN, "Undo session");

  rv=AQDB_DbFile_BeginEdit(db, AQDB_DBFILE_SESSION_FLAGS_ROLLBACK, NULL);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  DBG_INFO(AQDB_LOGDOMAIN, "Really undoing session");
  rv=AQDB_DbFile_ReallyUndoSession(db, sessionId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_EndEdit(db, AQDB_ACTION_FLAGS_ABORT);
    return rv;
  }

  /* let all types write their object lists and column lists */
  rv=AQDB_DbFile_FlushTypeData(db);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  /* end edit */
  rv=AQDB_DB_EndEdit(db, 0);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DB_EndEdit(db, AQDB_ACTION_FLAGS_ABORT);
    return rv;
  }

  return 0;
}





/* end */


















