/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqFinance (c) by 2015 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/






AQDB_OBJECT *AQDB_DbFile_Object_fromBuffer(AQDB_ID tableId,
                                           AQDB_ID objectId,
                                           uint32_t nfields,
                                           const uint8_t *pStart,
                                           uint32_t bsize) {
  AQDB_OBJECT *o;
  uint32_t j;
  const uint8_t *p;
  const uint8_t *pEnd;

  assert(nfields);

  o=AQDB_Object_new(tableId, objectId, nfields);
  assert(o);

#if 0
  DBG_ERROR(AQDB_LOGDOMAIN, "Reading object from this buffer (%d bytes)", bsize);
  GWEN_Text_LogString((const char*)pStart, bsize, AQDB_LOGDOMAIN, GWEN_LoggerLevel_Error);
#endif

  /* scan buffer and set fields accordingly */
  p=pStart;
  pEnd=p+bsize;
  if (pEnd<p) {
    DBG_INFO(AQDB_LOGDOMAIN, "Empty buffer");
    AQDB_Object_free(o);
    return NULL;
  }

  j=0;
  while((p<pEnd) && (j<nfields)) {
    uint32_t len;
    uint8_t c;

    c=*(p++);
    switch(c) {
    case 0xf1:
      len=*(p++);
      break;

    case 0xf2:
      len=*(p++);
      len+=(*(p++))<<8;
      break;

    case 0xf3:
      len=*(p++);
      len+=(*(p++))<<8;
      len+=(*(p++))<<16;
      break;

    case 0xf4:
      len=*(p++);
      len+=(*(p++))<<8;
      len+=(*(p++))<<16;
      len+=(*(p++))<<24;
      break;

    case 0xf5:
    case 0xf6:
    case 0xf7:
    case 0xf8:
    case 0xf9:
    case 0xfa:
    case 0xfb:
    case 0xfc:
    case 0xfd:
    case 0xfe:
    case 0xff:
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid field size specifier for field (%d)", j);
      AQDB_Object_free(o);
      return NULL;

    default:
      len=c;
      len+=(*(p++))<<8;
      break;
    }

#if 0
    DBG_ERROR(AQDB_LOGDOMAIN, "Setting field %d (out of %d):", j, nfields);
    GWEN_Text_LogString((const char*)p, len, AQDB_LOGDOMAIN, GWEN_LoggerLevel_Error);
#endif

    if ((p+len)>(pEnd+1)) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid field size (buffer overflow)");
      AQDB_Object_free(o);
      return NULL;
    }
    AQDB_Object_SetField(o, j, p, len);
    p+=len;
    j++;
  }
  if (p!=pEnd) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Incomplete data in buffer (%d, %d)",
              (int) (p-pStart), (int) bsize);
    AQDB_Object_free(o);
    return NULL;
  }

  return o;
}



int AQDB_DbFile_Object_toBuffer(const AQDB_OBJECT *o, GWEN_BUFFER *buf) {
  uint32_t fieldCount;
  uint32_t j;

  fieldCount=AQDB_Object_GetFieldCount(o);
  for (j=0; j<fieldCount; j++) {
    uint8_t *pData;
    uint32_t len;
    int rv;
    uint8_t lh[5];

    rv=AQDB_Object_GetField(o, j, &pData, &len);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }

    if (len>0xffffff) {
      lh[0]=0xf4;
      *((uint32_t*)(lh+1))=htole32(len);
      GWEN_Buffer_AppendBytes(buf, (const char*) lh, 5);
    }
    else if (len>0xffff) {
      lh[0]=0xf3;
      *((uint32_t*)(lh+1))=htole32(len);
      GWEN_Buffer_AppendBytes(buf, (const char*) lh, 4);
    }
    else if (len>0xf000) {
      lh[0]=0xf2;
      *((uint16_t*)(lh+1))=htole16(len);
      GWEN_Buffer_AppendBytes(buf, (const char*) lh, 3);
    }
    else {
      *((uint16_t*)(lh))=htole16(len);
      GWEN_Buffer_AppendBytes(buf, (const char*) lh, 2);
    }

    if (len>0)
      GWEN_Buffer_AppendBytes(buf, (const char*) pData, len);
  }

  return 0;
}



int AQDB_DbFile_ReadObjectAt(AQDB_DB *db,
                             AQDB_OBJECT *o,
                             uint64_t offset,
                             uint64_t bitFields) {
  AQDB_DB_FILE *xdb;
  uint64_t bits;
  uint32_t j;
  int64_t bytesLeft;
  AQFS_ITEM_HEADER *ih;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  bits=bitFields;
  if (bits==0)
    bits=~bits;

  /* read header for object */
  ih=AQFS_ItemHeader_new();
  rv=AQFS_File_ReadItemHeaderAt(xdb->file, offset, ih);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQFS_ItemHeader_free(ih);
    return rv;
  }

  /* get offset for data and size of object from item header */
  bytesLeft=AQFS_ItemHeader_GetSizeUsed(ih);
  offset=AQFS_ItemHeader_GetOffsetDataStart(ih);
  AQFS_ItemHeader_free(ih);

  if (bytesLeft==0) {
    DBG_INFO(AQDB_LOGDOMAIN, "Empty buffer");
    return GWEN_ERROR_NO_DATA;
  }

  /* read every field for which the corresponding bit is set */
  j=0;
  while(bytesLeft>0) {
    uint8_t buffer[5];
    int64_t rd;
    uint64_t len;
    uint8_t c;
    uint8_t *p;

    /* read field header */
    p=buffer;
    rv=AQFS_File_ReadForcedAt(xdb->file, offset, p, (bytesLeft>5)?5:bytesLeft);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      return rv;
    }

    /* read length */
    c=*(p++);
    offset++;
    bytesLeft--;
    switch(c) {
    case 0xf1:
      if (bytesLeft<1) {
        DBG_ERROR(AQDB_LOGDOMAIN, "Too few bytes for field (%d)", j);
        return GWEN_ERROR_BAD_DATA;
      }
      len=*(p++);
      offset++;
      bytesLeft--;
      break;

    case 0xf2:
      if (bytesLeft<2) {
        DBG_ERROR(AQDB_LOGDOMAIN, "Too few bytes for field (%d)", j);
        return GWEN_ERROR_BAD_DATA;
      }
      len=le16toh(*((uint16_t*)(p)));
      bytesLeft-=2;
      p+=2;
      break;

    case 0xf3:
      if (bytesLeft<3) {
        DBG_ERROR(AQDB_LOGDOMAIN, "Too few bytes for field (%d)", j);
        return GWEN_ERROR_BAD_DATA;
      }
      p[3]=0;
      len=le32toh(*((uint32_t*)(p)));
      offset+=3;
      bytesLeft-=3;
      p+=3;
      break;

    case 0xf4:
      if (bytesLeft<4) {
        DBG_ERROR(AQDB_LOGDOMAIN, "Too few bytes for field (%d)", j);
        return GWEN_ERROR_BAD_DATA;
      }
      len=le32toh(*((uint32_t*)(p)));
      offset+=4;
      bytesLeft-=4;
      p+=4;
      break;

    case 0xf5:
    case 0xf6:
    case 0xf7:
    case 0xf8:
    case 0xf9:
    case 0xfa:
    case 0xfb:
    case 0xfc:
    case 0xfd:
    case 0xfe:
    case 0xff:
      DBG_ERROR(AQDB_LOGDOMAIN, "Invalid field size specifier for field (%d)", j);
      return GWEN_ERROR_BAD_DATA;

    default:
      /* 16 bits */
      if (bytesLeft<1) {
        DBG_ERROR(AQDB_LOGDOMAIN, "Too few bytes for field (%d)", j);
        return GWEN_ERROR_BAD_DATA;
      }
      len=le16toh(*((uint16_t*)(p-1)));
      offset++;
      bytesLeft--;
      break;
    }

    /* now len contains the size of the data field, offset points to the data area */
    if (len) {
      if (len>bytesLeft) {
        DBG_ERROR(AQDB_LOGDOMAIN, "Too few bytes in object (%d > %d)", (int) len, (int) bytesLeft);
        return GWEN_ERROR_BAD_DATA;
      }

      if (bits & 1) {
        uint8_t *m;

        m=(uint8_t*) malloc(len);
        rd=AQFS_File_ReadAt(xdb->file, offset, m, len);
        if (rd<1) {
          DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
          free(m);
          return (int) rd;
        }
        AQDB_Object_SetField(o, j, m, len);
        free(m); /* TODO: omit copying here */
      }
    }

    bytesLeft-=len;
    offset+=len;
    j++;
    bits>>=1;
  } /* while */

  return 0;
}



int AQDB_DbFile_WriteObjectAt(AQDB_DB *db,
                              AQDB_OBJECT *o,
                              uint64_t offset) {
  AQDB_DB_FILE *xdb;
  GWEN_BUFFER *tbuf;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);


  tbuf=GWEN_Buffer_new(0, 256, 0, 1);
  rv=AQDB_DbFile_Object_toBuffer(o, tbuf);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_Buffer_free(tbuf);
    return rv;
  }

  rv=AQFS_File_WriteDataAt(xdb->file, offset, AQFS_ITEMHEADER_TYPE_GENERIC_DATA,
                           (const uint8_t*) GWEN_Buffer_GetStart(tbuf),
                           GWEN_Buffer_GetUsedBytes(tbuf));
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_Buffer_free(tbuf);
    return rv;
  }

  GWEN_Buffer_free(tbuf);
  return 0;
}



int64_t AQDB_DbFile_AppendObject(AQDB_DB *db,
                                 AQFS_TABLE_LIST *ftl,
                                 AQDB_OBJECT *o) {
  AQDB_DB_FILE *xdb;
  GWEN_BUFFER *tbuf;
  int rv;
  int64_t rd;
  uint64_t offset;
  uint64_t len;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  /* write object to a writable buffer */
  tbuf=GWEN_Buffer_new(0, 256, 0, 1);
  rv=AQDB_DbFile_Object_toBuffer(o, tbuf);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_Buffer_free(tbuf);
    return rv;
  }
  len=(uint64_t) GWEN_Buffer_GetUsedBytes(tbuf);

  /* append data to file */
  rd=AQFS_File_AppendData(xdb->file, AQFS_ITEMHEADER_TYPE_GENERIC_DATA,
                          (const uint8_t*) GWEN_Buffer_GetStart(tbuf),
                          len);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_Buffer_free(tbuf);
    return rd;
  }
  GWEN_Buffer_free(tbuf);
  offset=rd;

  /* add an entry for the data object to the given table list */
  rd=AQFS_File_AddTableEntry(xdb->file, ftl, offset);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rd;
  }
  DBG_VERBOUS(AQDB_LOGDOMAIN, "Added object %u at %llu (%llu bytes)",
              (unsigned int) rd,
              (unsigned long long int) offset,
              (unsigned long long int) len);

  /* return index of new object into table list */
  return rd;
}



int AQDB_DbFile_ReallyAddObject(AQDB_DB *db,
                                AQDB_ID tableId,
                                AQDB_OBJECT *object) {
  AQDB_DB_FILE *xdb;
  AQDB_DBFILE_TABLE *dbt;
  AQFS_TABLE_LIST *ftl;
  int64_t rd;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  dbt=AQDB_DbFile_GetTableById(db, tableId);
  if (dbt==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table [%d] not found", (int) tableId);
    return GWEN_ERROR_NOT_FOUND;
  }

  ftl=AQDB_DbFile_Table_GetTableListObjects(dbt);
  if (ftl==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table [%d] has no object list", (int) tableId);
    return GWEN_ERROR_NOT_FOUND;
  }

  rd=AQDB_DbFile_AppendObject(db, ftl, object);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (error %d, table [%d])", (int) rd,  (int) tableId);
    return (int) rd;
  }

  AQDB_Object_SetObjectId(object, rd+1);
  AQDB_Object_SetTableId(object, tableId);

  DBG_VERBOUS(AQDB_LOGDOMAIN, "Added object %llu to table [%s]",
              (unsigned long long) (rd+1), AQDB_DbFile_Table_GetName(dbt));

  return 0;
}



int AQDB_DbFile_ReallyDeleteObject(AQDB_DB *db, AQDB_ID tableId, AQDB_ID objectId) {
  AQDB_DB_FILE *xdb;
  AQDB_DBFILE_TABLE *dbt;
  AQFS_TABLE_LIST *ftl;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_WRITE)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for writing");
    return GWEN_ERROR_NOT_OPEN;
  }

  if (objectId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid object id 0");
    return GWEN_ERROR_INVALID;
  }

  if (tableId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id 0");
    return GWEN_ERROR_INVALID;
  }

  dbt=AQDB_DbFile_GetTableById(db, tableId);
  if (dbt==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table [%d] not found", (int) tableId);
    return GWEN_ERROR_NOT_FOUND;
  }

  ftl=AQDB_DbFile_Table_GetTableListObjects(dbt);
  if (ftl==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table [%d] has no object list", (int) tableId);
    return GWEN_ERROR_NOT_FOUND;
  }

  /* modify the data object entry to in the given table list */
  AQFS_File_SetTableEntry(xdb->file, ftl, objectId-1, 0xffffffffffffffffL);
  DBG_INFO(AQDB_LOGDOMAIN, "Deleted object %u", (unsigned int) objectId-1);

  return 0;
}



int AQDB_DbFile_ReallyWriteObject(AQDB_DB *db, const AQDB_OBJECT *object) {
  AQDB_DB_FILE *xdb;
  AQDB_DBFILE_TABLE *dbt;
  AQFS_TABLE_LIST *ftl;
  uint64_t tableId;
  uint64_t objectId;
  GWEN_BUFFER *tbuf;
  int rv;
  int64_t rd;
  uint64_t offset;
  uint64_t len;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (!(xdb->openFlags & AQDB_ACTION_FLAGS_WRITE)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Database is not open for writing");
    return GWEN_ERROR_NOT_OPEN;
  }

  objectId=AQDB_Object_GetObjectId(object);
  if (objectId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid object id 0");
    return GWEN_ERROR_INVALID;
  }

  tableId=AQDB_Object_GetTableId(object);
  if (tableId==0) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Invalid table id 0");
    return GWEN_ERROR_INVALID;
  }

  dbt=AQDB_DbFile_GetTableById(db, tableId);
  if (dbt==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table [%d] not found", (int) tableId);
    return GWEN_ERROR_NOT_FOUND;
  }

  ftl=AQDB_DbFile_Table_GetTableListObjects(dbt);
  if (ftl==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table [%d] has no object list", (int) tableId);
    return GWEN_ERROR_NOT_FOUND;
  }

  /* write object to a writable buffer */
  tbuf=GWEN_Buffer_new(0, 256, 0, 1);
  rv=AQDB_DbFile_Object_toBuffer(object, tbuf);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_Buffer_free(tbuf);
    return rv;
  }
  len=(uint64_t) GWEN_Buffer_GetUsedBytes(tbuf);

  /* append data to file.
   * Objects are never overwritten. The new version of an object is just appended
   * to the data file and its entry in the object table is modified to reflect the
   * new position. As long as the object table list isn't written to the file the
   * database remains consistent.
   */
  rd=AQFS_File_AppendData(xdb->file, AQFS_ITEMHEADER_TYPE_GENERIC_DATA,
                          (const uint8_t*) GWEN_Buffer_GetStart(tbuf),
                          len);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    GWEN_Buffer_free(tbuf);
    return rd;
  }
  GWEN_Buffer_free(tbuf);
  offset=rd;

  /* modify the data object entry to in the given table list */
  AQFS_File_SetTableEntry(xdb->file, ftl, objectId-1, offset);
  DBG_VERBOUS(AQDB_LOGDOMAIN, "Written object %u at %llu (%llu bytes)",
              (unsigned int) objectId-1,
              (unsigned long long int) offset,
              (unsigned long long int) len);

  /* remove from cache */
  AQDB_DB_CachePurgeObject(db, tableId, objectId);

  return 0;
}



int AQDB_DbFile_ReallyReadObject(AQDB_DB *db,
                                 uint32_t aflags,
                                 AQDB_ID tableId,
                                 AQDB_ID objectId,
                                 AQDB_OBJECT **pObject) {
  AQDB_DB_FILE *xdb;
  AQDB_DBFILE_TABLE *dbt;
  AQFS_TABLE_LIST *ftl;
  int64_t rd;
  uint64_t offset;
  uint64_t len;
  uint8_t *ptr;
  AQDB_OBJECT *o;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  dbt=AQDB_DbFile_GetTableById(db, tableId);
  if (dbt==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table [%d] not found", (int) tableId);
    return GWEN_ERROR_NOT_FOUND;
  }

  DBG_VERBOUS(AQDB_LOGDOMAIN, "Reading object %d from table %s",
           (int) objectId, AQDB_DbFile_Table_GetName(dbt));

  DBG_VERBOUS(AQDB_LOGDOMAIN, "- table %d is \"%s\"",
              (int) tableId, AQDB_DbFile_Table_GetName(dbt));

  /* check cache */
  if (xdb->fileLock==NULL || (aflags & AQDB_ACCESS_FLAGS_ISWRITER)) {
    int rv;

    /* either not in session mode OR (in session mode AND caller is the writer), look in the cache */
    rv=AQDB_DB_CacheGetObject(db, tableId, objectId, &o);
    if (rv>=0) {
      *pObject=o;
      return rv;
    }
  }

  if (xdb->fileLock && !(aflags & AQDB_ACCESS_FLAGS_ISWRITER)) {
    DBG_VERBOUS(AQDB_LOGDOMAIN, "- Returning object from saved table");
    ftl=AQDB_DbFile_Table_GetSavedTableListObjects(dbt);
  }
  else {
    DBG_VERBOUS(AQDB_LOGDOMAIN, "- Returning object from normal table");
    ftl=AQDB_DbFile_Table_GetTableListObjects(dbt);
  }
  if (ftl==NULL) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table [%d] has no object list", (int) tableId);
    return GWEN_ERROR_NOT_FOUND;
  }

  /* get offset of the given object inside the file */
  rd=AQFS_File_GetTableEntry(xdb->file, ftl, objectId-1);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
    return (int) rd;
  }
  offset=rd;

  /* read object data */
  rd=AQFS_File_ReadDataAt(xdb->file, offset, &ptr, 0);
  if (rd<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", (int) rd);
    return (int) rd;
  }
  len=rd;

  /* create object from buffer */
  o=AQDB_DbFile_Object_fromBuffer(tableId, objectId,
                                  AQDB_DbFile_Table_GetColumnCount(dbt),
                                  ptr, len);
  if (o==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "here");
    free(ptr);
    return GWEN_ERROR_BAD_DATA;
  }
  free(ptr);

  /* possibly store in cache */
  if (xdb->fileLock==NULL || (aflags & AQDB_ACCESS_FLAGS_ISWRITER)) {
    int rv;

    /* either not in session mode OR (in session mode AND caller is the writer), look in the cache */
    rv=AQDB_DB_CacheSetObject(db, o, len);
    if (rv<0) {
      DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      AQDB_Object_free(o);
      return rv;
    }
  }

  *pObject=o;
  return 0;
}









