/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqFinance (c) by 2015 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/


#ifndef AQDB_DBFILE_P_H
#define AQDB_DBFILE_P_H


#include "dbfile_l.h"
#include "table_l.h"
#include "header_l.h"
#include "changelog.h"

#include <aqdatabase/aqdb_db.h>
#include <aqdatabase/aqdb_query_result.h>

#include <aqfilestorage/aqfs_file.h>
#include <aqfilestorage/aqfs_table.h>

#include <gwenhywfar/fslock.h>


#include <stdio.h>



#define AQDB_DBFILE_TYPETABLE_MAXENTRIES   128
#define AQDB_DBFILE_OBJECTTABLE_MAXENTRIES 1024
#define AQDB_DBFILE_COLUMNTABLE_MAXENTRIES 64


#define AQDB_DBFILE_DBDATA_INDEX_HEADER 0
#define AQDB_DBFILE_DBDATA_INDEX_TYPES  1


#define AQDB_DBFILE_TYPENAME_SESSIONS  "+SessionDescr"
#define AQDB_DBFILE_TYPENAME_CHANGELOG "+ChangeLog"


#define AQDB_DBFILE_SESSION_FLAGS_ROLLBACK       0x00000001




typedef struct AQDB_DB_FILE AQDB_DB_FILE;
struct AQDB_DB_FILE {
  GWEN_FSLOCK *fileLock;
  char *dataFileName;

  AQFS_FILE *file;

  uint64_t currentSessionId;
  AQDB_SESSION_DESCR *currentSession;

  uint64_t lastQueryId;
  AQDB_QUERY_RESULT_LIST *queryResultList;

  char *currentUserName;
  uint32_t openFlags;

  uint32_t editFlags;

  AQDB_DBFILE_HEADER *header;
  AQDB_ID tableIdSessionDescr;
  AQDB_ID tableIdChangeLog;

  /**
   * The first object in a database file is a table containing two objects for now:
   * - an object containing database information data
   * - an object containing a list of type tables
   */
  AQFS_TABLE_LIST *tableListDatabase;


  AQFS_TABLE_LIST *tableListTypes;
  AQDB_DBFILE_TABLE_LIST *tables;

};


void GWENHYWFAR_CB AQDB_DbFile_FreeData(void *bp, void *p);



int GWENHYWFAR_CB AQDB_DbFile_Create(AQDB_DB *db, uint32_t flags);

int GWENHYWFAR_CB AQDB_DbFile_Open(AQDB_DB *db, uint32_t flags);

int GWENHYWFAR_CB AQDB_DbFile_Close(AQDB_DB *db, uint32_t flags);


int GWENHYWFAR_CB AQDB_DbFile_CreateTable(AQDB_DB *db,
                                          const char *name,
                                          uint32_t flags,
                                          const AQDB_COLUMN_LIST *columns);

int GWENHYWFAR_CB AQDB_DbFile_OpenTable(AQDB_DB *db,
                                        uint32_t aflags,
                                        const char *tname,
                                        uint32_t flags,
                                        AQDB_ID *pTableId);

int GWENHYWFAR_CB AQDB_DbFile_CloseTable(AQDB_DB *db,
                                         AQDB_ID tableId,
                                         uint32_t flags);

int GWENHYWFAR_CB AQDB_DbFile_AddObject(AQDB_DB *db,
                                        AQDB_ID tableId,
                                        AQDB_OBJECT *object);

int GWENHYWFAR_CB AQDB_DbFile_WriteObject(AQDB_DB *db, const AQDB_OBJECT *object);

int GWENHYWFAR_CB AQDB_DbFile_ReadObject(AQDB_DB *db,
                                         uint32_t aflags,
                                         AQDB_ID tableId,
                                         AQDB_ID objectId,
                                         AQDB_OBJECT **pObject);

int GWENHYWFAR_CB AQDB_DbFile_DeleteObject(AQDB_DB *db, AQDB_ID tableId, AQDB_ID objectId);


int GWENHYWFAR_CB AQDB_DbFile_BeginEdit(AQDB_DB *db, uint32_t flags, const char *user);

int GWENHYWFAR_CB AQDB_DbFile_EndEdit(AQDB_DB *db, uint32_t flags);

int GWENHYWFAR_CB AQDB_DbFile_QuerySubmit(AQDB_DB *db,
                                          uint32_t aflags,
                                          AQDB_ID tableId,
                                          AQDB_ID refQueryId,
                                          const char *queryExpression,
                                          uint32_t flags,
                                          AQDB_ID *pQueryId);

int GWENHYWFAR_CB AQDB_DbFile_QueryGetFirst(AQDB_DB *db,
                                            AQDB_ID queryId,
                                            AQDB_ID *pIdListPtr,
                                            int *pIdListLen);

int GWENHYWFAR_CB AQDB_DbFile_QueryGetNext(AQDB_DB *db,
                                           AQDB_ID queryId,
                                           AQDB_ID *pIdListPtr,
                                           int *pIdListLen);

int GWENHYWFAR_CB AQDB_DbFile_QueryClose(AQDB_DB *db, AQDB_ID queryId);


AQDB_COLUMN_LIST GWENHYWFAR_CB *AQDB_DbFile_GetColumnList(AQDB_DB *db,
                                                          uint32_t aflags,
                                                          AQDB_ID tableId);

int GWENHYWFAR_CB AQDB_DbFile_GetRollbackSessions(AQDB_DB *db,
                                                  uint32_t aflags,
                                                  AQDB_SESSION_DESCR_LIST *sl);

int GWENHYWFAR_CB AQDB_DbFile_UndoSession(AQDB_DB *db, AQDB_ID sessionId);







const char *AQDB_DbFile_GetCurrentUserName(const AQDB_DB *db);
void AQDB_DbFile_SetCurrentUserName(AQDB_DB *db, const char *userName);




int AQDB_DbFile_ReadTableAt(AQDB_DB *db, AQDB_DBFILE_TABLE *dbt, uint64_t offset);
int AQDB_DbFile_ReadTableFields(AQDB_DB *db, AQDB_DBFILE_TABLE *dbt);


AQDB_DBFILE_TABLE *AQDB_DbFile_GetTableById(AQDB_DB *db, AQDB_ID id);
AQDB_DBFILE_TABLE *AQDB_DbFile_GetTableByName(AQDB_DB *db, const char *name);


int AQDB_DbFile_ReadHeader(AQDB_DB *db, AQDB_DBFILE_HEADER *dbh);
int AQDB_DbFile_WriteHeader(AQDB_DB *db, AQDB_DBFILE_HEADER *dbh);

int AQDB_DbFile_CreateDatabaseStart(AQDB_DB *db);
int AQDB_DbFile_ReadDatabaseStart(AQDB_DB *db);

int AQDB_DbFile_ReadTypes(AQDB_DB *db);

int AQDB_DbFile_ReallyCreateTable(AQDB_DB *db,
                                  const char *name,
                                  uint32_t flags,
                                  const AQDB_COLUMN_LIST *columns);

int AQDB_DbFile_ReallyOpenTable(AQDB_DB *db,
                                uint32_t aflags,
                                const char *tname,
                                uint32_t flags,
                                AQDB_ID *pTableId);

int AQDB_DbFile_ReallyAddObject(AQDB_DB *db,
                                AQDB_ID tableId,
                                AQDB_OBJECT *object);

int AQDB_DbFile_ReallyDeleteObject(AQDB_DB *db, AQDB_ID tableId, AQDB_ID objectId);

int AQDB_DbFile_ReallyWriteObject(AQDB_DB *db, const AQDB_OBJECT *object);


int AQDB_DbFile_ReallyReadObject(AQDB_DB *db,
                                 uint32_t aflags,
                                 AQDB_ID tableId,
                                 AQDB_ID objectId,
                                 AQDB_OBJECT **pObject);

int AQDB_DbFile_ReallyQuerySubmit(AQDB_DB *db,
                                  uint32_t aflags,
                                  AQDB_ID tableId,
                                  AQDB_ID refQueryId,
                                  const char *queryExpression,
                                  uint32_t flags,
                                  AQDB_ID *pQueryId);

int AQDB_DbFile_ReallyQueryGetFirst(AQDB_DB *db,
                                    AQDB_ID queryId,
                                    AQDB_ID *pIdListPtr,
                                    int *pIdListLen);

int AQDB_DbFile_ReallyQueryGetNext(AQDB_DB *db,
                                   AQDB_ID queryId,
                                   AQDB_ID *pIdListPtr,
                                   int *pIdListLen);

int AQDB_DbFile_ReallyQueryClose(AQDB_DB *db, AQDB_ID queryId);

int AQDB_DbFile_ReallyUndoSession(AQDB_DB *db, AQDB_ID sessionId);


int AQDB_DbFile_GetMatchingObjects(AQDB_DB *db, AQDB_ID tableId,
                                   uint32_t aflags,
                                   const char *expr,
                                   uint32_t qflags,
                                   AQDB_OBJECT_LIST *ol);

int AQDB_DbFile_GetSessionDescriptions(AQDB_DB *db,
                                       uint32_t aflags,
                                       AQDB_ID minimumSessionId,
                                       AQDB_SESSION_DESCR_LIST *sl);

int AQDB_DbFile_GetChangelogsForSession(AQDB_DB *db,
                                        AQDB_ID sessionId,
                                        AQDB_DBFILE_CHANGELOG_LIST *cl);

int AQDB_DbFile_UndoSingleSession(AQDB_DB *db, const AQDB_SESSION_DESCR *sd);


int AQDB_DbFile_CloseSession(AQDB_DB *db);


void AQDB_DbFile_ReleaseInternalStructures(AQDB_DB *db);
int AQDB_DbFile_FlushTypeData(AQDB_DB *db);


#endif
