/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2016 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/




int AQDB_DbFile_ReallyQuerySubmit(AQDB_DB *db,
                                  uint32_t aflags,
                                  AQDB_ID tableId,
                                  AQDB_ID refQueryId,
                                  const char *queryExpression,
                                  uint32_t flags,
                                  AQDB_ID *pQueryId) {
  AQDB_DB_FILE *xdb;
  AQDB_DBFILE_TABLE *dbt=NULL;
  GWEN_IDLIST64 *idl=NULL;
  AQDB_QUERY_RESULT *qr;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (tableId==0) {
    DBG_INFO(AQDB_LOGDOMAIN, "Invalid table id");
    return GWEN_ERROR_INVALID;
  }
  dbt=AQDB_DbFile_GetTableById(db, tableId);
  if (dbt==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "No table with id %d", (int) tableId);
    return GWEN_ERROR_INVALID;
  }

  DBG_INFO(AQDB_LOGDOMAIN, "Submit query for table %s: [%s]",
           AQDB_DbFile_Table_GetName(dbt),
           queryExpression?queryExpression:"-empty-");

  if (!(AQDB_DbFile_Table_GetOpenFlags(dbt) & AQDB_ACTION_FLAGS_READ)) {
    DBG_ERROR(AQDB_LOGDOMAIN, "Table [%s] is not open",
              AQDB_DbFile_Table_GetName(dbt));
    return GWEN_ERROR_PERMISSIONS;
  }

  /* fill id list (either list of available ids or result of a previous query) */
  if (1) {
    if (refQueryId) { /* get results from previous query */
      AQDB_QUERY_RESULT *qrRef;

      qrRef=AQDB_QueryResult_List_GetById(xdb->queryResultList, refQueryId);
      if (qrRef==NULL) {
        DBG_INFO(AQDB_LOGDOMAIN, "No active query with id %d", (int) refQueryId);
        return GWEN_ERROR_INVALID;
      }
      idl=AQDB_QueryResult_GetIdList(qrRef);
      assert(idl);
      GWEN_IdList64_Attach(idl);
    }
    else { /* get list of active ids */
      AQFS_TABLE_LIST *ftl;

      /* create a new id list */
      idl=GWEN_IdList64_new();
      if (xdb->fileLock && !(aflags & AQDB_ACCESS_FLAGS_ISWRITER)) {
        DBG_INFO(AQDB_LOGDOMAIN, "In edit mode and caller is not the writer");
        ftl=AQDB_DbFile_Table_GetSavedTableListObjects(dbt);
      }
      else {
        DBG_INFO(AQDB_LOGDOMAIN, "Not in edit mode or caller is the writer");
        ftl=AQDB_DbFile_Table_GetTableListObjects(dbt);
      }
      if (ftl) {
        AQFS_TABLE *ft;
        uint64_t currentTableId=0;

        /* read entries from all tables */
        ft=AQFS_Table_List_First(ftl);
        while(ft) {
          uint64_t maxEntries;
          uint64_t idx;
          uint64_t *ptr;
      
          maxEntries=AQFS_Table_GetMaxEntries(ft);
          DBG_VERBOUS(AQDB_LOGDOMAIN, "Reading object entries from table \"%s\" (maxEntries=%d, free entries=%d)",
                      AQDB_DbFile_Table_GetName(dbt),
                      (int) maxEntries,
                      (int) AQFS_Table_GetFreeEntries(ft));
          ptr=AQFS_Table_GetPtrEntries(ft);
          for (idx=0; idx<maxEntries; idx++) {
            uint64_t offset;
      
            offset=*(ptr++);
            if (offset>0 && offset !=0xffffffffffffffffL) {
              /* add id */
              GWEN_IdList64_AddId(idl, currentTableId+idx+1);
            }
          }

          currentTableId+=maxEntries;
          ft=AQFS_Table_List_Next(ft);
        }
      }
    }
  }


  /* create query result */
  qr=AQDB_QueryResult_new(++(xdb->lastQueryId));
  assert(qr);


  /* create a query from expression (if any) */
  if (queryExpression) {
    AQDB_COLUMN_LIST *cl;
    AQDB_QUERY *dbq;
    GWEN_IDLIST64_ITERATOR *it;
    uint64_t id;

    cl=AQDB_DbFile_Table_GetColumnList(dbt);
    if (cl==NULL || AQDB_Column_List_GetCount(cl)==0) {
      DBG_ERROR(AQDB_LOGDOMAIN, "Table has no columns");
      AQDB_QueryResult_free(qr);
      return GWEN_ERROR_INTERNAL;
    }

    /* parse expression */
    dbq=AQDB_ObjectQuery_new(queryExpression, cl);
    if (dbq==NULL) {
      DBG_INFO(AQDB_LOGDOMAIN, "No query generated");
      AQDB_QueryResult_free(qr);
      return GWEN_ERROR_INVALID;
    }

    it=GWEN_IdList64_Iterator_new(idl);
    id=GWEN_IdList64_Iterator_GetFirstId(it);
    while(id) {
      AQDB_OBJECT *o=NULL;
      int rv;

      rv=AQDB_DbFile_ReallyReadObject(db, aflags, tableId, id, &o);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
      }
      else {
        /* an object has been read, check whether it matches */
        rv=AQDB_ObjectQuery_Check(dbq, o);
        AQDB_Object_free(o);
        if (rv) {
          /* it actually matches, add it to the idlist */
#if 0
          DBG_ERROR(0, "Entry matches for query %08x [%s] (ref: %08x)", i+1,
                    queryExpression, refQueryId);
#endif
          AQDB_QueryResult_AddId(qr, id);
          if (flags & AQDB_QUERY_FLAGS_STOPATFIRST)
            break;
        }
#if 0
        else {
          /* nah, doesn't match, ignore */
          DBG_ERROR(0, "Entry %d does not match for query %08x [%s] (ref: %08x)",
                    (int) id,
                    i+1,
                    queryExpression, refQueryId);
        }
#endif
      }

      /* try next id */
      id=GWEN_IdList64_Iterator_GetNextId(it);
    }
    GWEN_IdList64_Iterator_free(it);
    GWEN_IdList64_free(idl);
  }
  else { /* just add all entries from the current id list to the QueryResult object */
    if (flags & AQDB_QUERY_FLAGS_STOPATFIRST) {
      /* just add the first id to the list */
      AQDB_QueryResult_AddId(qr, GWEN_IdList64_GetIdAt(idl, 0));
    }
    else {
      /* add all ids to the list */
      AQDB_QueryResult_SetIdList(qr, idl);
    }
  }


  /* check for results */
  if (GWEN_IdList64_GetEntryCount(AQDB_QueryResult_GetIdList(qr))==0) {
    /* empty, discard result list */
    AQDB_QueryResult_free(qr);
    return GWEN_ERROR_NOT_FOUND;
  }


  /* at least one result, add to query result list */
  AQDB_QueryResult_List_Add(qr, xdb->queryResultList);
  if (pQueryId)
    *pQueryId=AQDB_QueryResult_GetQueryId(qr);
  return 0;
}



int AQDB_DbFile_ReallyQueryGetFirst(AQDB_DB *db,
                                    AQDB_ID queryId,
                                    AQDB_ID *pIdListPtr,
                                    int *pIdListLen) {
  AQDB_DB_FILE *xdb;
  int rv;
  AQDB_QUERY_RESULT *qr;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (queryId==0) {
    DBG_INFO(AQDB_LOGDOMAIN, "Invalid query id %d", queryId);
    return GWEN_ERROR_INVALID;
  }

  qr=AQDB_QueryResult_List_GetById(xdb->queryResultList, queryId);
  if (qr==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "Query %d not in use", (int) queryId);
    return GWEN_ERROR_INVALID;
  }

  rv=AQDB_QueryResult_GetFirstIds(qr, pIdListPtr, pIdListLen);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_DbFile_ReallyQueryGetNext(AQDB_DB *db,
                                   AQDB_ID queryId,
                                   AQDB_ID *pIdListPtr,
                                   int *pIdListLen) {
  AQDB_DB_FILE *xdb;
  int rv;
  AQDB_QUERY_RESULT *qr;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (queryId==0) {
    DBG_INFO(AQDB_LOGDOMAIN, "Invalid query id %d", queryId);
    return GWEN_ERROR_INVALID;
  }

  qr=AQDB_QueryResult_List_GetById(xdb->queryResultList, queryId);
  if (qr==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "Query %d not in use", (int) queryId);
    return GWEN_ERROR_INVALID;
  }


  rv=AQDB_QueryResult_GetNextIds(qr, pIdListPtr, pIdListLen);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}



int AQDB_DbFile_ReallyQueryClose(AQDB_DB *db, AQDB_ID queryId) {
  AQDB_DB_FILE *xdb;
  AQDB_QUERY_RESULT *qr;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  if (queryId==0) {
    DBG_INFO(AQDB_LOGDOMAIN, "Invalid query id %d", queryId);
    return GWEN_ERROR_INVALID;
  }

  qr=AQDB_QueryResult_List_GetById(xdb->queryResultList, queryId);
  if (qr==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "Query %d not in use", (int) queryId);
    return GWEN_ERROR_INVALID;
  }

  AQDB_QueryResult_List_Del(qr);
  AQDB_QueryResult_free(qr);

  return 0;
}




int AQDB_DbFile_GetMatchingObjects(AQDB_DB *db, AQDB_ID tableId,
                                   uint32_t aflags,
                                   const char *expr,
                                   uint32_t qflags,
                                   AQDB_OBJECT_LIST *ol) {
  AQDB_DB_FILE *xdb;
  AQDB_ID queryId;
  AQDB_ID resultList[64];
  int idListLen;
  int i;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  rv=AQDB_DbFile_ReallyQuerySubmit(db, aflags, tableId, 0, expr, qflags, &queryId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  idListLen=64;
  rv=AQDB_DbFile_ReallyQueryGetFirst(db, queryId, resultList, &idListLen);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DbFile_ReallyQueryClose(db, queryId);
    return rv;
  }

  for (;;) {
    for (i=0; i<idListLen; i++) {
      AQDB_OBJECT *o=NULL;

      rv=AQDB_DbFile_ReallyReadObject(db, aflags, tableId, resultList[i], &o);
      if (rv<0) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        AQDB_DbFile_ReallyQueryClose(db, queryId);
        return rv;
      }
      assert(o);
      AQDB_Object_List_Add(o, ol);
    }

    idListLen=64;
    rv=AQDB_DbFile_ReallyQueryGetNext(db, queryId, resultList, &idListLen);
    if (rv<0) {
      if (rv!=GWEN_ERROR_NOT_FOUND) {
        DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
        AQDB_DbFile_ReallyQueryClose(db, queryId);
        return rv;
      }
      else
        break;
    }
  }

  rv=AQDB_DbFile_ReallyQueryClose(db, queryId);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    return rv;
  }

  return 0;
}




