/****************************************************************************
 * This file is part of the project AqDataBase.
 * AqDatabase (c) by 2016 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/



int AQDB_DbFile_ReallyGetRollbackSessions(AQDB_DB *db,
                                          uint32_t aflags,
                                          AQDB_SESSION_DESCR_LIST *sl) {
  AQDB_DB_FILE *xdb;
  AQDB_DBFILE_TABLE *dbt;
  AQFS_TABLE_LIST *ftl;
  int numOfSessions=0;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  dbt=AQDB_DbFile_GetTableById(db, xdb->tableIdSessionDescr);
  if (dbt==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "Session description table not found, SNH!");
    return GWEN_ERROR_INTERNAL;
  }

  ftl=AQDB_DbFile_Table_GetTableListObjects(dbt);
  if (ftl) {
    AQFS_TABLE *ft;
    uint64_t currentTableId=0;

    /* read entries from all tables */
    ft=AQFS_Table_List_First(ftl);
    while(ft) {
      uint64_t maxEntries;
      uint64_t idx;
      uint64_t *ptr;
  
      maxEntries=AQFS_Table_GetMaxEntries(ft);
      DBG_VERBOUS(AQDB_LOGDOMAIN, "Reading object entries from table (maxEntries=%d, free entries=%d)",
                  (int) maxEntries,
                  (int) AQFS_Table_GetFreeEntries(ft));
      ptr=AQFS_Table_GetPtrEntries(ft);
      for (idx=0; idx<maxEntries; idx++) {
        uint64_t offset;
        uint64_t id;

        id=currentTableId+idx+1;
        offset=*(ptr++);
        if (offset>0 && offset !=0xffffffffffffffffL) {
          AQDB_OBJECT *o=NULL;
          int rv;
          AQDB_SESSION_DESCR *sd=NULL;

          /* read object from file */
          rv=AQDB_SessionDescr_CreateObject(db, xdb->tableIdSessionDescr, id, &o);
          if (rv<0) {
            DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
            return rv;
          }
          rv=AQDB_DbFile_ReadObjectAt(db, o, offset, 0);
          if (rv<0) {
            DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
            return rv;
          }

          /* create session descriptor */
          rv=AQDB_SessionDescr_fromObject(o, &sd);
          if (rv<0) {
            DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
            AQDB_Object_free(o);
            return rv;
	  }
	  AQDB_SessionDescr_SetDbId(sd, AQDB_Object_GetObjectId(o));
          AQDB_Object_free(o);

          AQDB_SessionDescr_List_Add(sd, sl);
          numOfSessions++;
        }
      }

      currentTableId+=maxEntries;
      ft=AQFS_Table_List_Next(ft);
    }
  }

  if (numOfSessions==0) {
    DBG_INFO(AQDB_LOGDOMAIN, "No sessions found");
    return GWEN_ERROR_NOT_FOUND;
  }

  return 0;
}



int AQDB_DbFile_GetSessionDescriptions(AQDB_DB *db,
                                       uint32_t aflags,
                                       AQDB_ID minimumSessionId,
                                       AQDB_SESSION_DESCR_LIST *sl) {
  AQDB_DB_FILE *xdb;
  AQDB_OBJECT_LIST *ol;
  char exprbuf[64];
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  ol=AQDB_Object_List_new();

  snprintf(exprbuf, sizeof(exprbuf)-1, "$sessionId>=\"%llu\"",
	   (long long unsigned int) minimumSessionId);
  exprbuf[sizeof(exprbuf)-1]=0;

  rv=AQDB_DbFile_GetMatchingObjects(db,
                                    xdb->tableIdSessionDescr,
                                    aflags, /* aflags */
				    exprbuf, /* expression */
				    0, /* query flags */
				    ol);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_List_free(ol);
    return rv;
  }

  if (AQDB_Object_List_GetCount(ol)) {
    AQDB_OBJECT *o;

    o=AQDB_Object_List_First(ol);
    while(o) {
      AQDB_SESSION_DESCR *sd=NULL;

      /* create change log entry */
      rv=AQDB_SessionDescr_fromObject(o, &sd);
      if (rv<0) {
	DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
	AQDB_Object_List_free(ol);
	return rv;
      }
      AQDB_SessionDescr_SetDbId(sd, AQDB_Object_GetObjectId(o));

      AQDB_SessionDescr_List_Add(sd, sl);
      o=AQDB_Object_List_Next(o);
    }
  }
  else {
    DBG_INFO(AQDB_LOGDOMAIN, "No session found >= %llu",
             (unsigned long long) minimumSessionId);
  }
  AQDB_Object_List_free(ol);

  return 0;
}



int AQDB_DbFile_GetChangelogsForSession(AQDB_DB *db,
					AQDB_ID sessionId,
					AQDB_DBFILE_CHANGELOG_LIST *cl) {
  AQDB_DB_FILE *xdb;
  AQDB_OBJECT_LIST *ol;
  char exprbuf[64];
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  ol=AQDB_Object_List_new();

  snprintf(exprbuf, sizeof(exprbuf)-1, "$sessionLogId==\"%llu\"",
	   (long long unsigned int) sessionId);
  exprbuf[sizeof(exprbuf)-1]=0;

  rv=AQDB_DbFile_GetMatchingObjects(db,
				    xdb->tableIdChangeLog,
				    AQDB_ACCESS_FLAGS_ISWRITER, /* aflags */
				    exprbuf, /* expression */
				    0, /* query flags */
				    ol);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_Object_List_free(ol);
    return rv;
  }

  if (AQDB_Object_List_GetCount(ol)) {
    AQDB_OBJECT *o;

    o=AQDB_Object_List_First(ol);
    while(o) {
      AQDB_DBFILE_CHANGELOG *c=NULL;

      /* create change log entry */
      rv=AQDB_DbFile_Changelog_fromObject(o, &c);
      if (rv<0) {
	DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
	AQDB_Object_List_free(ol);
	return rv;
      }
      AQDB_DbFile_Changelog_SetDbId(c, AQDB_Object_GetObjectId(o));

      AQDB_DbFile_Changelog_List_Add(c, cl);
      o=AQDB_Object_List_Next(o);
    }
  }
  AQDB_Object_List_free(ol);

  return 0;
}



int AQDB_DbFile_UndoSingleSession(AQDB_DB *db, const AQDB_SESSION_DESCR *sd) {
  AQDB_DB_FILE *xdb;
  AQDB_ID sessionId;
  AQDB_DBFILE_CHANGELOG_LIST *cl;
  int rv;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  sessionId=AQDB_SessionDescr_GetSessionId(sd);
  DBG_INFO(AQDB_LOGDOMAIN, "Undoing session %llu with %llu additions",
	    (unsigned long long) sessionId,
	   (unsigned long long) AQDB_SessionDescr_GetAdditions(sd));

  cl=AQDB_DbFile_Changelog_List_new();
  rv=AQDB_DbFile_GetChangelogsForSession(db, sessionId, cl);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_DbFile_Changelog_List_free(cl);
    return rv;
  }

  if (AQDB_DbFile_Changelog_List_GetCount(cl)) {
    AQDB_DBFILE_CHANGELOG *c;

    /* Sort backwards by database id */
    AQDB_DbFile_Changelog_List_SortByDbId(cl, 0);
    c=AQDB_DbFile_Changelog_List_First(cl);
    while(c) {
      AQDB_ID tableId;
      AQDB_ID objectId;
      AQDB_ID oldOffset;

      tableId=AQDB_DbFile_Changelog_GetTableId(c);
      objectId=AQDB_DbFile_Changelog_GetObjectId(c);
      oldOffset=AQDB_DbFile_Changelog_GetOldOffset(c);

      if (tableId && objectId) {
	AQDB_DBFILE_TABLE *dbt;
	AQFS_TABLE_LIST *ftl;

	dbt=AQDB_DbFile_GetTableById(db, tableId);
	if (dbt==NULL) {
	  DBG_INFO(AQDB_LOGDOMAIN, "Table %llu not found", (unsigned long long int) tableId);
	  return GWEN_ERROR_INTERNAL;
	}

	ftl=AQDB_DbFile_Table_GetTableListObjects(dbt);
	if (ftl==NULL) {
	  DBG_INFO(AQDB_LOGDOMAIN, "Table %llu has no file tablem SNH!", (unsigned long long int) tableId);
	  return GWEN_ERROR_INTERNAL;
	}

	switch(AQDB_DbFile_Changelog_GetEntryType(c)) {
	case AQDB_DbFile_ChangelogType_Unknown:
	case AQDB_DbFile_ChangelogType_None:
	  break;
	case AQDB_DbFile_ChangelogType_Addition:
	  /* object was added, delete the entry */
	  AQFS_File_SetTableEntry(xdb->file, ftl, objectId-1, 0xffffffffffffffffL);
	  break;
	case AQDB_DbFile_ChangelogType_Deletion:
	  /* object was deleted, add it back again */
	  AQFS_File_SetTableEntry(xdb->file, ftl, objectId-1, oldOffset);
	  break;
	case AQDB_DbFile_ChangelogType_Modification:
	  /* object was modified, set old entry */
	  AQFS_File_SetTableEntry(xdb->file, ftl, objectId-1, oldOffset);
	  break;
	}
      }

      c=AQDB_DbFile_Changelog_List_Next(c);
    }
  }

  return 0;
}


int AQDB_DbFile_ReallyUndoSession(AQDB_DB *db, AQDB_ID sessionId) {
  AQDB_DB_FILE *xdb;
  AQDB_SESSION_DESCR_LIST *sl;
  int rv;
  AQDB_DBFILE_TABLE *dbt;
  AQFS_TABLE_LIST *ftl;

  assert(db);
  xdb=GWEN_INHERIT_GETDATA(AQDB_DB, AQDB_DB_FILE, db);
  assert(xdb);

  /* get table info of the session descriptor table */
  dbt=AQDB_DbFile_GetTableById(db, xdb->tableIdSessionDescr);
  if (dbt==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "Session description table not found, SNH!");
    return GWEN_ERROR_INTERNAL;
  }

  ftl=AQDB_DbFile_Table_GetTableListObjects(dbt);
  if (ftl==NULL) {
    DBG_INFO(AQDB_LOGDOMAIN, "Session descriptor table has no file tablem SNH!");
    return GWEN_ERROR_INTERNAL;
  }

  /* read all rollback sessions */
  sl=AQDB_SessionDescr_List_new();
  rv=AQDB_DbFile_GetSessionDescriptions(db, AQDB_ACCESS_FLAGS_ISWRITER, sessionId, sl);
  if (rv<0) {
    DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
    AQDB_SessionDescr_List_free(sl);
    return rv;
  }

  if (AQDB_SessionDescr_List_GetCount(sl)) {
    AQDB_SESSION_DESCR *sd;

    /* make sure the order is correct */
    AQDB_SessionDescr_List_SortBySessionId(sl, 1);
    sd=AQDB_SessionDescr_List_Last(sl);
    while (sd) {
      AQDB_ID sid;

      sid=AQDB_SessionDescr_GetSessionId(sd);
      if (sid==sessionId) {
	DBG_INFO(AQDB_LOGDOMAIN, "Undoing session %llu",
		 (unsigned long long) AQDB_SessionDescr_GetSessionId(sd));
	rv=AQDB_DbFile_UndoSingleSession(db, sd);
	if (rv<0) {
	  DBG_INFO(AQDB_LOGDOMAIN, "here (%d)", rv);
	  AQDB_SessionDescr_List_free(sl);
	  return rv;
	}

	/* delete session entry from session descriptor list */
	AQFS_File_SetTableEntry(xdb->file, ftl, AQDB_SessionDescr_GetDbId(sd), 0xffffffffffffffffL);
      }
      sd=AQDB_SessionDescr_List_Previous(sd);
    }
  }
  else {
    DBG_INFO(AQDB_LOGDOMAIN, "No session found >= %llu",
             (unsigned long long) sessionId);
  }

  AQDB_SessionDescr_List_free(sl);
  return 0;
}


